# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from fastapi import FastAPI, HTTPException, Response
from fastapi.concurrency import run_in_threadpool
from fastapi.testclient import TestClient
from starlette.exceptions import HTTPException as StarletteHTTPException
from instana.span.span import get_current_span

fastapi_server = FastAPI()

# @fastapi_server.exception_handler(StarletteHTTPException)
# async def http_exception_handler(request, exc):
#     return PlainTextResponse(str(exc.detail), status_code=exc.status_code)

# @fastapi_server.exception_handler(RequestValidationError)
# async def validation_exception_handler(request, exc):
#     return PlainTextResponse(str(exc), status_code=400)


@fastapi_server.get("/")
async def root():
    return {"message": "Hello World"}


@fastapi_server.get("/users/{user_id}")
async def user(user_id):
    return {"user": user_id}


@fastapi_server.get("/response_headers")
async def response_headers():
    headers = {"X-Capture-This-Too": "this too", "X-Capture-That-Too": "that too"}
    return Response("Stan wuz here with headers!", headers=headers)


@fastapi_server.get("/400")
async def four_zero_zero():
    raise HTTPException(status_code=400, detail="400 response")


@fastapi_server.get("/404")
async def four_zero_four():
    raise HTTPException(status_code=404, detail="Item not found")


@fastapi_server.get("/500")
async def five_hundred():
    raise HTTPException(status_code=500, detail="500 response")


@fastapi_server.get("/starlette_exception")
async def starlette_exception():
    raise StarletteHTTPException(status_code=500, detail="500 response")


def trigger_outgoing_call():
    # As TestClient() is based on httpx, and we don't support it yet,
    # we must pass the SDK trace_id and span_id to the ASGI server.
    span_context = get_current_span().get_span_context()
    headers = {
        "X-INSTANA-T": str(span_context.trace_id),
        "X-INSTANA-S": str(span_context.span_id),
    }
    client = TestClient(fastapi_server, headers=headers)
    response = client.get("/users/1")
    return response.json()


@fastapi_server.get("/non_async_simple")
def non_async_complex_call():
    response = trigger_outgoing_call()
    return response


@fastapi_server.get("/non_async_threadpool")
def non_async_threadpool():
    run_in_threadpool(trigger_outgoing_call)
    return {
        "message": "non async functions executed on a thread pool can't be followed through thread boundaries"
    }



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

# This is a wrapper for PEP-0249: Python Database API Specification v2.0
import wrapt
from typing import TYPE_CHECKING, Dict, Any, List, Tuple, Union, Callable, Optional
from typing_extensions import Self

from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import SpanKind

from instana.log import logger
from instana.util.traceutils import get_tracer_tuple, tracing_is_off
from instana.util.sql import sql_sanitizer

if TYPE_CHECKING:
    from instana.span.span import InstanaSpan


class CursorWrapper(wrapt.ObjectProxy):
    __slots__ = ("_module_name", "_connect_params", "_cursor_params")

    def __init__(
        self,
        cursor: Any,
        module_name: str,
        connect_params: Optional[List[Union[str, Dict[str, Any]]]] = None,
        cursor_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        super(CursorWrapper, self).__init__(wrapped=cursor)
        self._module_name = module_name
        self._connect_params = connect_params
        self._cursor_params = cursor_params

    def _collect_kvs(
        self,
        span: "InstanaSpan",
        sql: str,
    ) -> None:
        try:
            span.set_attribute("span.kind", SpanKind.CLIENT)

            db_parameter_name = next(
                (
                    p
                    for p in ("db", "database", "dbname")
                    if p in self._connect_params[1]
                ),
                None,
            )
            if db_parameter_name:
                span.set_attribute(
                    SpanAttributes.DB_NAME,
                    self._connect_params[1][db_parameter_name],
                )

            span.set_attribute(SpanAttributes.DB_STATEMENT, sql_sanitizer(sql))
            span.set_attribute(SpanAttributes.DB_USER, self._connect_params[1]["user"])
            span.set_attribute("host", self._connect_params[1]["host"])
            span.set_attribute("port", self._connect_params[1]["port"])
        except Exception as e:
            logger.debug(e)

    def __enter__(self) -> Self:
        return self

    def execute(
        self,
        sql: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Callable[[str, Dict[str, Any]], None]:
        tracer, parent_span, operation_name = get_tracer_tuple()

        # If not tracing or we're being called from sqlalchemy, just pass through
        if tracing_is_off() or (operation_name == "sqlalchemy"):
            return self.__wrapped__.execute(sql, params)

        parent_context = parent_span.get_span_context() if parent_span else None
        with tracer.start_as_current_span(
            self._module_name, span_context=parent_context
        ) as span:
            try:
                self._collect_kvs(span, sql)
                result = self.__wrapped__.execute(sql, params)
            except Exception as e:
                if span:
                    span.record_exception(e)
                raise
            else:
                return result

    def executemany(
        self,
        sql: str,
        seq_of_parameters: List[Dict[str, Any]],
    ) -> Callable[[str, List[Dict[str, Any]]], None]:
        tracer, parent_span, operation_name = get_tracer_tuple()

        # If not tracing or we're being called from sqlalchemy, just pass through
        if tracing_is_off() or (operation_name == "sqlalchemy"):
            return self.__wrapped__.executemany(sql, seq_of_parameters)

        parent_context = parent_span.get_span_context() if parent_span else None
        with tracer.start_as_current_span(
            self._module_name, span_context=parent_context
        ) as span:
            try:
                self._collect_kvs(span, sql)
                result = self.__wrapped__.executemany(sql, seq_of_parameters)
            except Exception as e:
                if span:
                    span.record_exception(e)
                raise
            else:
                return result

    def callproc(
        self,
        proc_name: str,
        params: Dict[str, Any],
    ) -> Callable[[str, Dict[str, Any]], None]:
        tracer, parent_span, operation_name = get_tracer_tuple()

        # If not tracing or we're being called from sqlalchemy, just pass through
        if tracing_is_off() or (operation_name == "sqlalchemy"):
            return self.__wrapped__.execute(proc_name, params)

        parent_context = parent_span.get_span_context() if parent_span else None
        with tracer.start_as_current_span(
            self._module_name, span_context=parent_context
        ) as span:
            try:
                self._collect_kvs(span, proc_name)
                result = self.__wrapped__.callproc(proc_name, params)
            except Exception:
                try:
                    result = self.__wrapped__.execute(proc_name, params)
                except Exception as e_execute:
                    if span:
                        span.record_exception(e_execute)
                    raise
                else:
                    return result
            else:
                return result


class ConnectionWrapper(wrapt.ObjectProxy):
    __slots__ = ("_module_name", "_connect_params")

    def __init__(
        self,
        connection: "ConnectionWrapper",
        module_name: str,
        connect_params: List[Union[str, Dict[str, Any]]],
    ) -> None:
        super(ConnectionWrapper, self).__init__(wrapped=connection)
        self._module_name = module_name
        self._connect_params = connect_params

    def __enter__(self) -> Self:
        return self

    def cursor(
        self,
        *args: Tuple[int, str, Dict[str, Any]],
        **kwargs: Dict[str, Any],
    ) -> CursorWrapper:
        return CursorWrapper(
            cursor=self.__wrapped__.cursor(*args, **kwargs),
            module_name=self._module_name,
            connect_params=self._connect_params,
            cursor_params=(args, kwargs) if args or kwargs else None,
        )

    def close(self) -> Callable[[], None]:
        return self.__wrapped__.close()

    def commit(self) -> Callable[[], None]:
        return self.__wrapped__.commit()

    def rollback(self) -> Callable[[], None]:
        return self.__wrapped__.rollback()


class ConnectionFactory(object):
    def __init__(
        self,
        connect_func: CursorWrapper,
        module_name: str,
    ) -> None:
        self._connect_func = connect_func
        self._module_name = module_name
        self._wrapper_ctor = ConnectionWrapper

    def __call__(
        self,
        *args: Tuple[int, str, Dict[str, Any]],
        **kwargs: Dict[str, Any],
    ) -> ConnectionWrapper:
        connect_params = (args, kwargs) if args or kwargs else None
        return self._wrapper_ctor(
            connection=self._connect_func(*args, **kwargs),
            module_name=self._module_name,
            connect_params=connect_params,
        )



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019

"""
The uwsgi and uwsgidecorators packages are added automatically to the Python environment
when running under uWSGI.  Here we attempt to detect the presence of these packages and
then use the appropriate hooks.
"""

try:
    from instana.log import logger
    from instana.singletons import agent

    import uwsgi

    logger.debug(
        f"uWSGI options: {uwsgi.opt}",
    )

    opt_master = uwsgi.opt.get("master", False)
    opt_lazy_apps = uwsgi.opt.get("lazy-apps", False)

    if not uwsgi.opt.get("enable-threads", False) and not uwsgi.opt.get(
        "gevent", False
    ):
        logger.warning(
            "Required: Neither uWSGI threads or gevent is enabled. "
            + "Please enable by using the uWSGI --enable-threads or --gevent option."
        )

    if opt_master and not opt_lazy_apps:
        # --master is supplied in uWSGI options (otherwise uwsgidecorators package won't be available)
        # When --lazy-apps is True, this postfork hook isn't needed
        import uwsgidecorators

        @uwsgidecorators.postfork
        def uwsgi_handle_fork() -> None:
            """This is our uWSGI hook to detect and act when worker processes are forked off."""
            logger.debug("Handling uWSGI fork...")
            agent.handle_fork()

        logger.debug("Applied uWSGI hooks")
    else:
        logger.debug(
            f"uWSGI --master={opt_master} --lazy-apps={opt_lazy_apps}: postfork hooks not applied"
        )
except ImportError:
    logger.debug(
        "uwsgi hooks: decorators not available: likely not running under uWSGI"
    )
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
from typing import Generator

import pytest

from instana import autoprofile
from instana.autoprofile.profiler import Profiler


class TestFrameCache:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Create a new Profiler.
        self.profiler = Profiler(None)
        self.profiler.start(disable_timers=True)
        yield
        # teardown
        self.profiler.destroy()

    def test_skip_stack(self) -> None:
        test_profiler_file = os.path.realpath(autoprofile.__file__)

        assert self.profiler.frame_cache.is_profiler_frame(test_profiler_file)



# coding=utf-8
# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016
"""
Instana

https://www.ibm.com/products/instana

Documentation: https://www.ibm.com/docs/en/instana-observability/current
Source Code: https://github.com/instana/python-sensor
"""

import importlib
import os
import sys
from typing import Tuple

from instana.collector.helpers.runtime import (
    is_autowrapt_instrumented,
    is_webhook_instrumented,
)
from instana.version import VERSION

__author__ = "Instana Inc."
__copyright__ = "Copyright 2020 Instana Inc."
__credits__ = ["Pavlo Baron", "Peter Giacomo Lombardo", "Andrey Slotin"]
__license__ = "MIT"
__maintainer__ = "Peter Giacomo Lombardo"
__email__ = "peter.lombardo@instana.com"
__version__ = VERSION

# User configurable EUM API key for instana.helpers.eum_snippet()
# pylint: disable=invalid-name
eum_api_key = ""

# This Python package can be loaded into Python processes one of three ways:
#   1. manual import statement
#   2. autowrapt hook
#   3. dynamically injected remotely
#
# With such magic, we may get pulled into Python processes that we have no interest being in.
# As a safety measure, we maintain a "do not load list" and if this process matches something
# in that list, then we go sit in a corner quietly and don't load anything at all.
do_not_load_list = [
    "pip",
    "pip2",
    "pip3",
    "pipenv",
    "docker-compose",
    "easy_install",
    "easy_install-2.7",
    "smtpd.py",
    "twine",
    "ufw",
    "unattended-upgrade",
]


def load(_: object) -> None:
    """
    Method used to activate the Instana sensor via AUTOWRAPT_BOOTSTRAP
    environment variable.
    """
    # Work around https://bugs.python.org/issue32573
    if not hasattr(sys, "argv"):
        sys.argv = [""]
    return None


def apply_gevent_monkey_patch() -> None:
    from gevent import monkey

    if os.environ.get("INSTANA_GEVENT_MONKEY_OPTIONS"):

        def short_key(k: str) -> str:
            return k[3:] if k.startswith("no-") else k

        def key_to_bool(k: str) -> bool:
            return not k.startswith("no-")

        import inspect

        all_accepted_patch_all_args = inspect.getfullargspec(monkey.patch_all)[0]
        provided_options = (
            os.environ.get("INSTANA_GEVENT_MONKEY_OPTIONS")
            .replace(" ", "")
            .replace("--", "")
            .split(",")
        )
        provided_options = [
            k for k in provided_options if short_key(k) in all_accepted_patch_all_args
        ]

        fargs = {
            short_key(k): key_to_bool(k)
            for (k, v) in zip(provided_options, [True] * len(provided_options))
        }
        monkey.patch_all(**fargs)
    else:
        monkey.patch_all()


def get_aws_lambda_handler() -> Tuple[str, str]:
    """
    For instrumenting AWS Lambda, users specify their original lambda handler
    in the LAMBDA_HANDLER environment variable.  This function searches for and
    parses that environment variable or returns the defaults.

    The default handler value for AWS Lambda is 'lambda_function.lambda_handler'
    which equates to the function "lambda_handler in a file named
    lambda_function.py" or in Python terms
    "from lambda_function import lambda_handler"
    """
    handler_module = "lambda_function"
    handler_function = "lambda_handler"

    try:
        handler = os.environ.get("LAMBDA_HANDLER", False)

        if handler:
            parts = handler.split(".")
            handler_function = parts.pop().strip()
            handler_module = ".".join(parts).strip()
    except Exception as exc:
        print(f"get_aws_lambda_handler error: {exc}")

    return handler_module, handler_function


def lambda_handler(event: str, context: str) -> None:
    """
    Entry point for AWS Lambda monitoring.

    This function will trigger the initialization of Instana monitoring and then call
    the original user specified lambda handler function.
    """
    module_name, function_name = get_aws_lambda_handler()

    try:
        # Import the module specified in module_name
        handler_module = importlib.import_module(module_name)
    except ImportError:
        print(
            f"Couldn't determine and locate default module handler: {module_name}.{function_name}"
        )
    else:
        # Now get the function and execute it
        if hasattr(handler_module, function_name):
            handler_function = getattr(handler_module, function_name)
            return handler_function(event, context)
        else:
            print(
                f"Couldn't determine and locate default function handler: {module_name}.{function_name}"
            )


def boot_agent() -> None:
    """Initialize the Instana agent and conditionally load auto-instrumentation."""

    import instana.singletons  # noqa: F401

    # Instrumentation
    if "INSTANA_DISABLE_AUTO_INSTR" not in os.environ:
        # TODO: remove the following entries as the migration of the
        # instrumentation codes are finalised.

        # Import & initialize instrumentation
        from instana.instrumentation import (
            asyncio,  # noqa: F401
            boto3_inst,  # noqa: F401
            cassandra_inst,  # noqa: F401
            couchbase_inst,  # noqa: F401
            fastapi_inst,  # noqa: F401
            flask,  # noqa: F401
            # gevent_inst,  # noqa: F401
            grpcio,  # noqa: F401
            logging,  # noqa: F401
            mysqlclient,  # noqa: F401
            pika,  # noqa: F401
            pep0249,  # noqa: F401
            psycopg2,  # noqa: F401
            pymongo,  # noqa: F401
            pymysql,  # noqa: F401
            pyramid,  # noqa: F401
            redis,  # noqa: F401
            sqlalchemy,  # noqa: F401
            starlette_inst,  # noqa: F401
            sanic_inst,  # noqa: F401
            urllib3,  # noqa: F401
        )
        from instana.instrumentation.aiohttp import (
            client,  # noqa: F401
            server,  # noqa: F401
        )
        from instana.instrumentation.aws import lambda_inst  # noqa: F401
        from instana.instrumentation import celery  # noqa: F401
        from instana.instrumentation.django import middleware  # noqa: F401
        from instana.instrumentation.google.cloud import (
            pubsub,  # noqa: F401
            storage,  # noqa: F401
        )
        from instana.instrumentation.tornado import (
            client,  # noqa: F401
            server,  # noqa: F401
        )

    # Hooks
    from instana.hooks import hook_uwsgi, hook_gunicorn  # noqa: F401


if "INSTANA_DISABLE" not in os.environ:
    # There are cases when sys.argv may not be defined at load time.  Seems to happen in embedded Python,
    # and some Pipenv installs.  If this is the case, it's best effort.
    if (
        hasattr(sys, "argv")
        and len(sys.argv) > 0
        and (os.path.basename(sys.argv[0]) in do_not_load_list)
    ):
        if "INSTANA_DEBUG" in os.environ:
            print(
                f"Instana: No use in monitoring this process type ({os.path.basename(sys.argv[0])}). Will go sit in a corner quietly."
            )
    else:
        # Automatic gevent monkey patching
        # unless auto instrumentation is off, then the customer should do manual gevent monkey patching
        if (
            (is_autowrapt_instrumented() or is_webhook_instrumented())
            and "INSTANA_DISABLE_AUTO_INSTR" not in os.environ
            and importlib.util.find_spec("gevent")
        ):
            apply_gevent_monkey_patch()
        # AutoProfile
        if "INSTANA_AUTOPROFILE" in os.environ:
            from instana.singletons import get_profiler

            profiler = get_profiler()
            if profiler:
                profiler.start()

        boot_agent()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

try:
    import sys

    from django import VERSION as django_version
    from opentelemetry import context, trace
    from opentelemetry.semconv.trace import SpanAttributes
    import wrapt
    from typing import TYPE_CHECKING, Dict, Any, Callable, Optional, List, Tuple, Type

    from instana.log import logger
    from instana.singletons import agent, tracer
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import extract_custom_headers
    from instana.propagators.format import Format

    if TYPE_CHECKING:
        from django.core.handlers.base import BaseHandler
        from django.http import HttpRequest, HttpResponse

    DJ_INSTANA_MIDDLEWARE = (
        "instana.instrumentation.django.middleware.InstanaMiddleware"
    )

    if django_version >= (2, 0):
        # Since Django 2.0, only `settings.MIDDLEWARE` is supported, so new-style
        # middlewares can be used.
        class MiddlewareMixin:
            def __init__(self, get_response):
                self.get_response = get_response

            def __call__(self, request):
                self.process_request(request)
                response = self.get_response(request)
                return self.process_response(request, response)

    else:
        # Note: For 1.11 <= django_version < 2.0
        # Django versions 1.x can use `settings.MIDDLEWARE_CLASSES` and expect
        # old-style middlewares, which are created by inheriting from
        # `deprecation.MiddlewareMixin` since its creation in Django 1.10 and 1.11
        from django.utils.deprecation import MiddlewareMixin

    class InstanaMiddleware(MiddlewareMixin):
        """Django Middleware to provide request tracing for Instana"""

        def __init__(
            self,
            get_response: Optional[Callable[["HttpRequest"], "HttpResponse"]] = None,
        ) -> None:
            super(InstanaMiddleware, self).__init__(get_response)
            self.get_response = get_response

        def process_request(self, request: Type["HttpRequest"]) -> None:
            try:
                env = request.META

                span_context = tracer.extract(Format.HTTP_HEADERS, env)

                span = tracer.start_span("django", span_context=span_context)
                request.span = span

                ctx = trace.set_span_in_context(span)
                token = context.attach(ctx)
                request.token = token

                extract_custom_headers(span, env, format=True)

                request.span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)
                if "PATH_INFO" in env:
                    request.span.set_attribute(
                        SpanAttributes.HTTP_URL, env["PATH_INFO"]
                    )
                if "QUERY_STRING" in env and len(env["QUERY_STRING"]):
                    scrubbed_params = strip_secrets_from_query(
                        env["QUERY_STRING"],
                        agent.options.secrets_matcher,
                        agent.options.secrets_list,
                    )
                    request.span.set_attribute("http.params", scrubbed_params)
                if "HTTP_HOST" in env:
                    request.span.set_attribute(SpanAttributes.HTTP_HOST, env["HTTP_HOST"])
            except Exception:
                logger.debug("Django middleware @ process_request", exc_info=True)

        def process_response(
            self, request: Type["HttpRequest"], response: "HttpResponse"
        ) -> "HttpResponse":
            try:
                if request.span:
                    if 500 <= response.status_code:
                        request.span.assure_errored()
                    # for django >= 2.2
                    if request.resolver_match is not None and hasattr(
                        request.resolver_match, "route"
                    ):
                        path_tpl = request.resolver_match.route
                    # django < 2.2 or in case of 404
                    else:
                        try:
                            from django.urls import resolve

                            view_name = resolve(request.path)._func_path
                            path_tpl = "".join(url_pattern_route(view_name))
                        except Exception:
                            # the resolve method can fire a Resolver404 exception, in this case there is no matching route
                            # so the path_tpl is set to None in order not to be added as a tag
                            path_tpl = None
                    if path_tpl:
                        request.span.set_attribute("http.path_tpl", path_tpl)

                    request.span.set_attribute(
                        SpanAttributes.HTTP_STATUS_CODE, response.status_code
                    )
                    if hasattr(response, "headers"):
                        extract_custom_headers(
                            request.span, response.headers, format=False
                        )
                    tracer.inject(request.span.context, Format.HTTP_HEADERS, response)
            except Exception:
                logger.debug("Instana middleware @ process_response", exc_info=True)
            finally:
                if hasattr(request, "span") and request.span:
                    if request.span.is_recording():
                        request.span.end()
                    request.span = None
                if hasattr(request, "token") and request.token:
                    context.detach(request.token)
                    request.token = None
            return response

        def process_exception(
            self, request: Type["HttpRequest"], exception: Exception
        ) -> None:
            from django.http.response import Http404

            if isinstance(exception, Http404):
                return None

            if request.span:
                request.span.record_exception(exception)

    def url_pattern_route(view_name: str) -> Callable[..., object]:
        from django.conf import settings

        try:
            from django.urls import (
                RegexURLPattern as URLPattern,
                RegexURLResolver as URLResolver,
            )
        except ImportError:
            from django.urls import URLPattern, URLResolver

        urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [""])

        def list_urls(
            urlpatterns: List[str], parent_pattern: Optional[List[str]] = None
        ) -> Callable[..., object]:
            if not urlpatterns:
                return
            if parent_pattern is None:
                parent_pattern = []
            first = urlpatterns[0]
            if isinstance(first, URLPattern):
                if first.lookup_str == view_name:
                    if hasattr(first, "regex"):
                        return parent_pattern + [str(first.regex.pattern)]
                    else:
                        return parent_pattern + [str(first.pattern)]
            elif isinstance(first, URLResolver):
                if hasattr(first, "regex"):
                    return list_urls(
                        first.url_patterns, parent_pattern + [str(first.regex.pattern)]
                    )
                else:
                    return list_urls(
                        first.url_patterns, parent_pattern + [str(first.pattern)]
                    )
            return list_urls(urlpatterns[1:], parent_pattern)

        return list_urls(urlconf.urlpatterns)

    def load_middleware_wrapper(
        wrapped: Callable[..., None],
        instance: Type["BaseHandler"],
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> Callable[..., None]:
        try:
            from django.conf import settings

            # Django >=1.10 to <2.0 support old-style MIDDLEWARE_CLASSES so we
            # do as well here
            if hasattr(settings, "MIDDLEWARE") and settings.MIDDLEWARE is not None:
                if DJ_INSTANA_MIDDLEWARE in settings.MIDDLEWARE:
                    return wrapped(*args, **kwargs)

                if isinstance(settings.MIDDLEWARE, tuple):
                    settings.MIDDLEWARE = (DJ_INSTANA_MIDDLEWARE,) + settings.MIDDLEWARE
                elif isinstance(settings.MIDDLEWARE, list):
                    settings.MIDDLEWARE = [DJ_INSTANA_MIDDLEWARE] + settings.MIDDLEWARE
                else:
                    logger.warning("Instana: Couldn't add InstanaMiddleware to Django")

            elif (
                hasattr(settings, "MIDDLEWARE_CLASSES")
                and settings.MIDDLEWARE_CLASSES is not None
            ):  # pragma: no cover
                if DJ_INSTANA_MIDDLEWARE in settings.MIDDLEWARE_CLASSES:
                    return wrapped(*args, **kwargs)

                if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                    settings.MIDDLEWARE_CLASSES = (
                        DJ_INSTANA_MIDDLEWARE,
                    ) + settings.MIDDLEWARE_CLASSES
                elif isinstance(settings.MIDDLEWARE_CLASSES, list):
                    settings.MIDDLEWARE_CLASSES = [
                        DJ_INSTANA_MIDDLEWARE
                    ] + settings.MIDDLEWARE_CLASSES
                else:
                    logger.warning("Instana: Couldn't add InstanaMiddleware to Django")

            else:  # pragma: no cover
                logger.warning("Instana: Couldn't find middleware settings")

            return wrapped(*args, **kwargs)
        except Exception:
            logger.warning(
                "Instana: Couldn't add InstanaMiddleware to Django: ", exc_info=True
            )

    try:
        logger.debug("Instrumenting django")
        wrapt.wrap_function_wrapper(
            "django.core.handlers.base",
            "BaseHandler.load_middleware",
            load_middleware_wrapper,
        )

        if "/tmp/.instana/python" in sys.path:  # pragma: no cover
            # If we are instrumenting via AutoTrace (in an already running process), then the
            # WSGI middleware has to be live reloaded.
            from django.core.servers.basehttp import get_internal_wsgi_application
            from django.core.exceptions import ImproperlyConfigured

            try:
                wsgiapp = get_internal_wsgi_application()
                wsgiapp.load_middleware()
            except ImproperlyConfigured:
                pass

    except Exception:
        logger.debug("django.middleware:", exc_info=True)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2024

from typing import TYPE_CHECKING, Type, List

from opentelemetry.trace.span import format_span_id
from opentelemetry.trace import SpanKind

from instana.util.ids import hex_id
if TYPE_CHECKING:
    from instana.span.base_span import BaseSpan


def format_span(
    queued_spans: List[Type["BaseSpan"]],
) -> List[Type["BaseSpan"]]:
    """
    Format Span Kind and the Trace, Parent Span and Span IDs of the Spans to be a 64-bit
    Hexadecimal String instead of Integer before being pushed to a
    Collector (or Instana Agent).
    """
    spans = []
    for span in queued_spans:
        span.t = format_span_id(span.t)
        span.s = format_span_id(span.s)
        span.p = format_span_id(span.p) if span.p else None
        span.lt = hex_id(span.lt) if hasattr(span, "lt") else None
        if isinstance(span.k, SpanKind):
            span.k = span.k.value if not span.k is SpanKind.INTERNAL else 3
        spans.append(span)
    return spans



# (c) Copyright IBM Corp. 2024

from typing import Tuple

from instana.span.base_span import BaseSpan
from instana.span.kind import ENTRY_KIND, EXIT_KIND
from instana.util import DictionaryOfStan


class SDKSpan(BaseSpan):
    def __init__(self, span, source, service_name, **kwargs) -> None:
        # pylint: disable=invalid-name
        super(SDKSpan, self).__init__(span, source, **kwargs)

        span_kind = self.get_span_kind(span)

        self.n = "sdk"
        self.k = span_kind[1]

        if service_name is not None:
            self.data["service"] = service_name

        self.data["sdk"]["name"] = span.name
        self.data["sdk"]["type"] = span_kind[0]
        self.data["sdk"]["custom"]["tags"] = self._validate_attributes(
            span.attributes
        )

        if span.events is not None and len(span.events) > 0:
            events = DictionaryOfStan()
            for event in span.events:
                filtered_attributes = self._validate_attributes(event.attributes)
                if len(filtered_attributes.keys()) > 0:
                    events[repr(event.timestamp)] = filtered_attributes
            self.data["sdk"]["custom"]["events"] = events

        if "arguments" in span.attributes:
            self.data["sdk"]["arguments"] = span.attributes["arguments"]

        if "return" in span.attributes:
            self.data["sdk"]["return"] = span.attributes["return"]

        # if len(span.context.baggage) > 0:
        #     self.data["baggage"] = span.context.baggage

    def get_span_kind(self, span) -> Tuple[str, int]:
        """
        Will retrieve the `span.kind` attribute and return a tuple containing the appropriate string and integer
        values for the Instana backend

        :param span: The span to search for the `span.kind` attribute
        :return: Tuple (String, Int)
        """
        kind = ("intermediate", 3)
        if "span.kind" in span.attributes:
            if span.attributes["span.kind"] in ENTRY_KIND:
                kind = ("entry", 1)
            elif span.attributes["span.kind"] in EXIT_KIND:
                kind = ("exit", 2)
        return kind





# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import wrapt
import flask
from importlib.metadata import version
from typing import Callable, Tuple, Dict, Any, TYPE_CHECKING, Union

from opentelemetry.semconv.trace import SpanAttributes

from instana.log import logger
from instana.singletons import tracer
from instana.propagators.format import Format


if TYPE_CHECKING:
    from werkzeug.exceptions import HTTPException
    from flask.typing import ResponseReturnValue
    from jinja2.environment import Template

@wrapt.patch_function_wrapper('flask', 'templating._render')
def render_with_instana(
    wrapped: Callable[..., str],
    instance: object,
    argv: Tuple[flask.app.Flask, "Template", Dict[str, Any]],
    kwargs: Dict[str, Any],
) -> str:
    # If we're not tracing, just return
    if not (hasattr(flask, "g") and hasattr(flask.g, "span")):
        return wrapped(*argv, **kwargs)

    parent_span = flask.g.span
    parent_context = parent_span.get_span_context()

    with tracer.start_as_current_span("render", span_context=parent_context) as span:
        try:
            flask_version = tuple(map(int, version("flask").split(".")))
            template = argv[1] if flask_version >= (2, 2, 0) else argv[0]

            span.set_attribute("type", "template")
            if template.name is None:
                span.set_attribute("name", "(from string)")
            else:
                span.set_attribute("name", template.name)

            return wrapped(*argv, **kwargs)
        except Exception as e:
            span.record_exception(e)
            raise


@wrapt.patch_function_wrapper('flask', 'Flask.handle_user_exception')
def handle_user_exception_with_instana(
    wrapped: Callable[..., Union["HTTPException", "ResponseReturnValue"]],
    instance: flask.app.Flask,
    argv: Tuple[Exception],
    kwargs: Dict[str, Any],
) -> Union["HTTPException", "ResponseReturnValue"]:
    # Call original and then try to do post processing
    response = wrapped(*argv, **kwargs)

    try:
        exc = argv[0]

        if hasattr(flask.g, "span") and flask.g.span:
            span = flask.g.span

            if response:
                if isinstance(response, tuple):
                    status_code = response[1]
                else:
                    if hasattr(response, 'code'):
                        status_code = response.code
                    else:
                        status_code = response.status_code

                if 500 <= status_code:
                    span.record_exception(exc)

                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, int(status_code))

                if hasattr(response, 'headers'):
                    tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)
            if span and span.is_recording():
                span.end()
            flask.g.span = None
    except:
        logger.debug("handle_user_exception_with_instana:", exc_info=True)

    return response



import time
import pytest
from typing import Generator
from instana.collector.utils import format_span
from instana.singletons import tracer
from instana.span.registered_span import RegisteredSpan
from instana.span.span import InstanaSpan, get_current_span
from opentelemetry.trace.span import format_span_id
from opentelemetry.trace import SpanKind

from instana.span_context import SpanContext


class TestUtils:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.span_context = None
        yield

    def test_format_span(self, span_context: SpanContext) -> None:
        self.span_context = span_context
        with tracer.start_as_current_span(
            name="span1", span_context=self.span_context
        ) as pspan:
            expected_trace_id = format_span_id(pspan.context.trace_id)
            expected_span_id = format_span_id(pspan.context.span_id)
            assert get_current_span() is pspan
            with tracer.start_as_current_span(name="span2") as cspan:
                assert get_current_span() is cspan
                assert cspan.parent_id == pspan.context.span_id
                span_list = [
                    RegisteredSpan(pspan, None, "log"),
                    RegisteredSpan(cspan, None, "log"),
                ]
        formatted_spans = format_span(span_list)
        assert len(formatted_spans) == 2
        assert formatted_spans[0].t == expected_trace_id
        assert formatted_spans[0].k == 1
        assert formatted_spans[0].s == expected_span_id
        assert formatted_spans[0].n == "span1"

        assert formatted_spans[1].t == expected_trace_id
        assert formatted_spans[1].p == formatted_spans[0].s
        assert formatted_spans[1].k == 1
        assert formatted_spans[1].s != formatted_spans[0].s
        assert formatted_spans[1].n == "span2"



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import signal
from typing import TYPE_CHECKING

from instana.autoprofile.runtime import RuntimeInfo, register_signal

if TYPE_CHECKING:
    from types import FrameType


class TestRuntime:
    def test_register_signal(self) -> None:
        if RuntimeInfo.OS_WIN:
            return

        result = {"handler": 0}

        def _handler(signum: signal.Signals, frame: "FrameType") -> None:
            result["handler"] += 1

        register_signal(signal.SIGUSR1, _handler)

        os.kill(os.getpid(), signal.SIGUSR1)
        os.kill(os.getpid(), signal.SIGUSR1)

        signal.signal(signal.SIGUSR1, signal.SIG_DFL)

        assert result["handler"] == 2



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import os
from typing import Any, Dict, Generator

import pytest
from opentelemetry.trace import (
    INVALID_SPAN_ID,
    INVALID_TRACE_ID,
    format_span_id,
    format_trace_id,
)

from instana.propagators.http_propagator import HTTPPropagator
from instana.span_context import SpanContext
from instana.util.ids import header_to_long_id, internal_id


class TestHTTPPropagator:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        self.hptc = HTTPPropagator()
        yield
        # teardown
        # Clear the INSTANA_DISABLE_W3C_TRACE_CORRELATION environment variable
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = ""

    @pytest.fixture(scope="function")
    def _instana_long_tracer_id(self) -> str:
        return "4bf92f3577b34da6a3ce929d0e0e4736"

    @pytest.fixture(scope="function")
    def _instana_span_id(self) -> str:
        return "00f067aa0ba902b7"

    @pytest.fixture(scope="function")
    def _trace_id(self, _instana_long_tracer_id: str) -> int:
        return int(_instana_long_tracer_id[-16:], 16)

    @pytest.fixture(scope="function")
    def _span_id(self, _instana_span_id: str) -> int:
        return int(_instana_span_id, 16)

    @pytest.fixture(scope="function")
    def _long_tracer_id(self, _instana_long_tracer_id: str) -> int:
        return int(_instana_long_tracer_id, 16)

    @pytest.fixture(scope="function")
    def _traceparent(self, _instana_long_tracer_id: str, _instana_span_id: str) -> str:
        return f"00-{_instana_long_tracer_id}-{_instana_span_id}-01"

    @pytest.fixture(scope="function")
    def _tracestate(self) -> str:
        return "congo=t61rcWkgMzE"

    def test_extract_carrier_dict(
        self,
        trace_id: int,
        span_id: int,
        _instana_long_tracer_id: str,
        _instana_span_id: str,
        _trace_id: int,
        _span_id: int,
        _traceparent: str,
        _tracestate: str,
    ) -> None:
        carrier = {
            "traceparent": _traceparent,
            "tracestate": _tracestate,
            "X-INSTANA-T": f"{trace_id}",
            "X-INSTANA-S": f"{span_id}",
            "X-INSTANA-L": f"1, correlationType=web; correlationId={span_id}",
        }

        ctx = self.hptc.extract(carrier)

        assert ctx.correlation_id == str(span_id)
        assert ctx.correlation_type == "web"
        assert not ctx.instana_ancestor
        assert ctx.level == 1
        assert ctx.long_trace_id == header_to_long_id(_instana_long_tracer_id)
        assert ctx.span_id == _span_id
        assert not ctx.synthetic
        assert ctx.trace_id == _trace_id
        assert ctx.trace_parent
        assert ctx.traceparent == f"00-{_instana_long_tracer_id}-{_instana_span_id}-01"
        assert ctx.tracestate == _tracestate

    def test_extract_carrier_list(
        self,
        _trace_id: int,
        _span_id: int,
        _instana_long_tracer_id: str,
        _instana_span_id: str,
        _traceparent: str,
        _tracestate: str,
    ) -> None:
        _trace_id = str(_trace_id)
        carrier = [
            ("user-agent", "python-requests/2.23.0"),
            ("accept-encoding", "gzip, deflate"),
            ("accept", "*/*"),
            ("connection", "keep-alive"),
            ("traceparent", _traceparent),
            ("tracestate", _tracestate),
            ("X-INSTANA-T", f"{_trace_id}"),
            ("X-INSTANA-S", f"{_span_id}"),
            ("X-INSTANA-L", "1"),
        ]

        ctx = self.hptc.extract(carrier)

        assert not ctx.correlation_id
        assert not ctx.correlation_type
        assert not ctx.instana_ancestor
        assert ctx.level == 1
        assert not ctx.long_trace_id
        assert ctx.span_id == _span_id
        assert not ctx.synthetic
        assert ctx.trace_id == internal_id(_trace_id)
        assert not ctx.trace_parent
        assert ctx.traceparent == f"00-{_instana_long_tracer_id}-{_instana_span_id}-01"
        assert ctx.tracestate == _tracestate

    def test_extract_carrier_dict_validate_Exception_None_returned(
        self,
        trace_id: int,
        span_id: int,
        _tracestate: str,
    ) -> None:
        # In this test case, the traceparent header fails the validation, so
        # traceparent and tracestate are not used.
        # Additionally, because the correlation flags are present in the
        # 'X-INSTANA-L' header, we need to start a new SpanContext, and the
        # present values of 'X-INSTANA-T' and 'X-INSTANA-S' headers should not
        # be used.

        carrier = {
            "traceparent": "00-4gf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01'",  # the long-trace-id is malformed to be invalid.
            "tracestate": _tracestate,
            "X-INSTANA-T": f"{trace_id}",
            "X-INSTANA-S": f"{span_id}",
            "X-INSTANA-L": f"1, correlationType=web; correlationId={span_id}",
        }

        ctx = self.hptc.extract(carrier)

        assert isinstance(ctx, SpanContext)
        assert ctx.trace_id == INVALID_TRACE_ID
        assert ctx.span_id == INVALID_SPAN_ID
        assert not ctx.synthetic
        assert ctx.correlation_id == str(span_id)
        assert ctx.correlation_type == "web"

    def test_extract_fake_exception(
        self,
        trace_id: int,
        span_id: int,
        _tracestate: str,
        mocker,
    ) -> None:
        carrier = {
            "traceparent": "00-4bf92f3577b34da6a3ce929d0e0e-00f067aa0ba902b7-01",
            "tracestate": _tracestate,
            "X-INSTANA-T": f"{trace_id}",
            "X-INSTANA-S": f"{span_id}",
            "X-INSTANA-L": f"1, correlationType=web; correlationId={span_id}",
        }
        with pytest.raises(Exception):
            ctx = self.hptc.extract(carrier)
            assert not ctx

    def test_extract_carrier_dict_corrupted_level_header(
        self,
        trace_id: int,
        span_id: int,
        _instana_long_tracer_id: str,
        _trace_id: int,
        _span_id: int,
        _traceparent: str,
        _tracestate: str,
    ) -> None:
        # In this test case, the 'X-INSTANA-L' header is corrupted

        carrier = {
            "traceparent": _traceparent,
            "tracestate": _tracestate,
            "X-INSTANA-T": f"{trace_id}",
            "X-INSTANA-S": f"{span_id}",
            "X-INSTANA-L": f"1, correlationType=web; correlationId{span_id}",
        }

        ctx = self.hptc.extract(carrier)

        assert not ctx.correlation_id
        assert ctx.correlation_type == "web"
        assert not ctx.instana_ancestor
        assert ctx.level == 1
        assert ctx.long_trace_id == header_to_long_id(_instana_long_tracer_id)
        assert ctx.span_id == _span_id
        assert not ctx.synthetic
        assert ctx.trace_id == _trace_id
        assert ctx.trace_parent
        assert ctx.traceparent == _traceparent
        assert ctx.tracestate == _tracestate

    def test_extract_carrier_dict_level_header_not_splitable(
        self,
        _trace_id: int,
        _span_id: int,
        _traceparent: str,
        _tracestate: str,
    ) -> None:
        _trace_id = str(_trace_id)
        carrier = {
            "traceparent": _traceparent,
            "tracestate": _tracestate,
            "X-INSTANA-T": f"{_trace_id}",
            "X-INSTANA-S": f"{_span_id}",
            "X-INSTANA-L": ["1"],
        }

        ctx = self.hptc.extract(carrier)

        assert not ctx.correlation_id
        assert not ctx.correlation_type
        assert not ctx.instana_ancestor
        assert ctx.level == 1
        assert not ctx.long_trace_id
        assert ctx.span_id == _span_id
        assert not ctx.synthetic
        assert ctx.trace_id == internal_id(_trace_id)
        assert not ctx.trace_parent
        assert ctx.traceparent == _traceparent
        assert ctx.tracestate == _tracestate

    # The following tests are based on the test cases defined in the
    # tracer_compliance_test_cases.json file.
    #
    # Each line of the parametrize tuple correlates to a test case scenario:
    # - scenario 28: "Scenario/incoming headers": "w3c off, only X-INSTANA-L=0"
    # - scenario 29: "Scenario/incoming headers": "w3c off, X-INSTANA-L=0 plus -T and -S"
    # - scenario 30: "Scenario/incoming headers": "w3c off, X-INSTANA-L=0 plus traceparent"
    # - scenario 31: "Scenario/incoming headers": "w3c off, X-INSTANA-L=0 plus traceparent and tracestate",
    @pytest.mark.parametrize(
        "disable_w3c, carrier_header",
        [
            ("yes_please", {"X-INSTANA-L": "0"}),
            (
                "w3c_trace_correlation_stinks",
                {
                    "X-INSTANA-T": "11803532876627986230",
                    "X-INSTANA-S": "67667974448284343",
                    "X-INSTANA-L": "0",
                },
            ),
            (
                "w3c_trace_correlation_stinks",
                {
                    "traceparent": "00-0af7651916cd43dd8448eb211c80319c-b9c7c989f97918e1-01",
                    "X-INSTANA-L": "0",
                },
            ),
            (
                "w3c_trace_correlation_stinks",
                {
                    "traceparent": "00-0af7651916cd43dd8448eb211c80319c-b9c7c989f97918e1-01",
                    "tracestate": "congo=ucfJifl5GOE,rojo=00f067aa0ba902b7",
                    "X-INSTANA-L": "0",
                },
            ),
        ],
    )
    def test_w3c_off_x_instana_l_0(
        self,
        disable_w3c: str,
        carrier_header: Dict[str, Any],
        trace_id: int,
    ) -> None:
        os.environ["INSTANA_DISABLE_W3C_TRACE_CORRELATION"] = disable_w3c

        ctx = self.hptc.extract(carrier_header)

        # Assert the level is (zero) int, not str
        assert isinstance(ctx.level, int)
        assert ctx.level == 0

        # Assert the suppression is on
        assert ctx.suppression

        # Assert the rest of the attributes are on their default value
        assert ctx.trace_id == INVALID_TRACE_ID
        assert ctx.span_id == INVALID_SPAN_ID
        assert not ctx.synthetic
        assert not ctx.correlation_id
        assert not ctx.trace_parent
        assert not ctx.instana_ancestor
        assert not ctx.long_trace_id
        assert not ctx.correlation_type
        assert not ctx.correlation_id

        # Assert that the traceparent is propagated when it is enabled
        if "traceparent" in carrier_header.keys():
            assert ctx.traceparent
            tp_trace_id = header_to_long_id(carrier_header["traceparent"].split("-")[1])
        else:
            assert not ctx.traceparent
            tp_trace_id = ctx.trace_id

        # Assert that the tracestate is propagated when it is enabled
        if "tracestate" in carrier_header.keys():
            assert ctx.tracestate
        else:
            assert not ctx.tracestate

        # Simulate the side-effect of starting a span, getting a trace_id and span_id.
        # Actually, with OTel API using a Tuple to store the SpanContext info,
        # this will not change the values.
        ctx.trace_id = ctx.span_id = trace_id

        # Test propagation
        downstream_carrier = {}

        self.hptc.inject(ctx, downstream_carrier)

        # Assert the 'X-INSTANA-L' has been injected with the correct 0 value
        assert "X-INSTANA-L" in downstream_carrier
        assert downstream_carrier.get("X-INSTANA-L") == "0"

        assert "traceparent" in downstream_carrier
        assert (
            downstream_carrier.get("traceparent")
            == f"00-{format_trace_id(tp_trace_id)}-{format_span_id(ctx.span_id)}-00"
        )

        # Assert that the tracestate is propagated when it is enabled
        if "tracestate" in carrier_header.keys():
            assert "tracestate" in downstream_carrier
            assert carrier_header["tracestate"] == downstream_carrier["tracestate"]



# (c) Copyright IBM Corp. 2024

from opentelemetry.trace import SpanKind

ENTRY_KIND = ("entry", "server", "consumer", SpanKind.SERVER, SpanKind.CONSUMER)

EXIT_KIND = ("exit", "client", "producer", SpanKind.CLIENT, SpanKind.PRODUCER)

LOCAL_SPANS = ("asyncio", "render", SpanKind.INTERNAL)

HTTP_SPANS = (
    "aiohttp-client",
    "aiohttp-server",
    "django",
    "http",
    "tornado-client",
    "tornado-server",
    "urllib3",
    "wsgi",
    "asgi",
)

ENTRY_SPANS = (
    "aiohttp-server",
    "aws.lambda.entry",
    "celery-worker",
    "django",
    "wsgi",
    "rabbitmq",
    "rpc-server",
    "tornado-server",
    "gcps-consumer",
    "asgi",
)

EXIT_SPANS = (
    "aiohttp-client",
    "boto3",
    "cassandra",
    "celery-client",
    "couchbase",
    "log",
    "memcache",
    "mongo",
    "mysql",
    "postgres",
    "rabbitmq",
    "redis",
    "rpc-client",
    "sqlalchemy",
    "tornado-client",
    "urllib3",
    "pymongo",
    "gcs",
    "gcps-producer",
)

REGISTERED_SPANS = LOCAL_SPANS + ENTRY_SPANS + EXIT_SPANS



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2018

import os
from typing import TYPE_CHECKING, Type

from opentelemetry import trace

from instana.recorder import StanRecorder
from instana.tracer import InstanaTracerProvider
from instana.autoprofile.profiler import Profiler

if TYPE_CHECKING:
    from instana.agent.base import BaseAgent
    from instana.tracer import InstanaTracer

agent = None
tracer = None
profiler = None
span_recorder = None

# Detect the environment where we are running ahead of time
aws_env = os.environ.get("AWS_EXECUTION_ENV", "")
env_is_aws_fargate = aws_env == "AWS_ECS_FARGATE"
env_is_aws_eks_fargate = (
    os.environ.get("INSTANA_TRACER_ENVIRONMENT") == "AWS_EKS_FARGATE"
)
env_is_aws_lambda = "AWS_Lambda_" in aws_env
k_service = os.environ.get("K_SERVICE")
k_configuration = os.environ.get("K_CONFIGURATION")
k_revision = os.environ.get("K_REVISION")
instana_endpoint_url = os.environ.get("INSTANA_ENDPOINT_URL")
env_is_google_cloud_run = all(
    (k_service, k_configuration, k_revision, instana_endpoint_url)
)

if env_is_aws_lambda:
    from .agent.aws_lambda import AWSLambdaAgent
    from .recorder import StanRecorder

    agent = AWSLambdaAgent()
elif env_is_aws_fargate:
    from instana.agent.aws_fargate import AWSFargateAgent
    agent = AWSFargateAgent()
elif env_is_google_cloud_run:
    from instana.agent.google_cloud_run import GCRAgent
    agent = GCRAgent(
        service=k_service, configuration=k_configuration, revision=k_revision
    )
elif env_is_aws_eks_fargate:
    from instana.agent.aws_eks_fargate import EKSFargateAgent
    agent = EKSFargateAgent()
else:
    from instana.agent.host import HostAgent
    agent = HostAgent()
    profiler = Profiler(agent)
    

if agent:
    span_recorder = StanRecorder(agent)        


def get_agent() -> Type["BaseAgent"]:
    """
    Retrieve the globally configured agent
    @return: The Instana Agent singleton
    """
    global agent
    return agent


def set_agent(new_agent: Type["BaseAgent"]) -> None:
    """
    Set the global agent for the Instana package.  This is used for the
    test suite only currently.

    @param new_agent: agent to replace current singleton
    @return: None
    """
    global agent
    agent = new_agent


# The global OpenTelemetry compatible tracer used internally by
# this package.
provider = InstanaTracerProvider(span_processor=span_recorder, exporter=agent)

# Sets the global default tracer provider
trace.set_tracer_provider(provider)

# Creates a tracer from the global tracer provider
tracer = trace.get_tracer("instana.tracer")


def get_tracer() -> "InstanaTracer":
    """
    Retrieve the globally configured tracer
    @return: Tracer
    """
    global tracer
    return tracer


def set_tracer(new_tracer: "InstanaTracer") -> None:
    """
    Set the global tracer for the Instana package.  This is used for the
    test suite only currently.
    @param new_tracer: The new tracer to replace the singleton
    @return: None
    """
    global tracer
    tracer = new_tracer


def get_profiler() -> Profiler:
    """
    Retrieve the globally configured profiler
    @return: Profiler
    """
    global profiler
    return profiler


def set_profiler(new_profiler: Profiler):
    """
    Set the global profiler for the Instana package.  This is used for the
    test suite only currently.
    @param new_profiler: The new profiler to replace the singleton
    @return: None
    """
    global profiler
    profiler = new_profiler



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import sys
from typing import Generator
import json
import pytest
import requests
import io

from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from tests.test_utils import _TraceContextMixin
from opentelemetry.trace import SpanKind

from mock import patch, Mock
from six.moves import http_client

from google.cloud import storage
from google.api_core import iam, page_iterator
from google.auth.credentials import AnonymousCredentials


class TestGoogleCloudStorage(_TraceContextMixin):
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        yield
        agent.options.allow_exit_as_root = False

    @patch("requests.Session.request")
    def test_buckets_list(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#buckets", "items": []},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            buckets = client.list_buckets()
            for _ in buckets:
                pass

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.list"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"

    @patch("requests.Session.request")
    def test_buckets_list_as_root_exit_span(self, mock_requests: Mock) -> None:
        agent.options.allow_exit_as_root = True
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#buckets", "items": []},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        buckets = client.list_buckets()
        for _ in buckets:
            pass

        spans = self.recorder.queued_spans()

        assert len(spans) == 1

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.list"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"

    @patch("requests.Session.request")
    def test_buckets_insert(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.create_bucket("test bucket")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.insert"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_get(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.get_bucket("test bucket")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        assert gcs_span.t == test_span.t
        assert gcs_span.p == test_span.s

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.get"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_patch(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").patch()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.patch"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_update(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#bucket"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").update()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.update"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_get_iam_policy(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#policy"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").get_iam_policy()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.getIamPolicy"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_set_iam_policy(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#policy"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").set_iam_policy(iam.Policy())

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.setIamPolicy"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_test_iam_permissions(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#testIamPermissionsResponse"},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").test_iam_permissions("test-permission")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.testIamPermissions"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_lock_retention_policy(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={
                "kind": "storage#bucket",
                "metageneration": 1,
                "retentionPolicy": {"isLocked": False},
            },
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        bucket = client.bucket("test bucket")
        bucket.reload()

        with tracer.start_as_current_span("test"):
            bucket.lock_retention_policy()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.lockRetentionPolicy"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_buckets_delete(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response()

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").delete()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "buckets.delete"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_objects_compose(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("dest object").compose(
                [
                    storage.blob.Blob("object 1", "test bucket"),
                    storage.blob.Blob("object 2", "test bucket"),
                ]
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.compose"
        assert gcs_span.data["gcs"]["destinationBucket"] == "test bucket"
        assert gcs_span.data["gcs"]["destinationObject"] == "dest object"
        assert (
            gcs_span.data["gcs"]["sourceObjects"]
            == "test bucket/object 1,test bucket/object 2"
        )

    @patch("requests.Session.request")
    def test_objects_copy(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        bucket = client.bucket("src bucket")

        with tracer.start_as_current_span("test"):
            bucket.copy_blob(
                bucket.blob("src object"),
                client.bucket("dest bucket"),
                new_name="dest object",
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.copy"
        assert gcs_span.data["gcs"]["destinationBucket"] == "dest bucket"
        assert gcs_span.data["gcs"]["destinationObject"] == "dest object"
        assert gcs_span.data["gcs"]["sourceBucket"] == "src bucket"
        assert gcs_span.data["gcs"]["sourceObject"] == "src object"

    @patch("requests.Session.request")
    def test_objects_delete(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response()

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").delete()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.delete"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_objects_attrs(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").exists()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.attrs"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @pytest.mark.skipif(
        sys.version_info >= (3, 14),
        reason='Avoiding "Fatal Python error: Segmentation fault"',
    )
    @patch("requests.Session.request")
    def test_objects_get(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            content=b"CONTENT", status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").download_to_file(
                io.BytesIO(), raw_download=True
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.get"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_objects_insert(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").upload_from_string(
                "CONTENT"
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.insert"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_objects_list(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            blobs = client.bucket("test bucket").list_blobs()

            for _ in blobs:
                pass

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.list"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_objects_patch(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").patch()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.patch"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_objects_rewrite(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={
                "kind": "storage#rewriteResponse",
                "totalBytesRewritten": 0,
                "objectSize": 0,
                "done": True,
                "resource": {},
            },
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("dest bucket").blob("dest object").rewrite(
                client.bucket("src bucket").blob("src object")
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.rewrite"
        assert gcs_span.data["gcs"]["destinationBucket"] == "dest bucket"
        assert gcs_span.data["gcs"]["destinationObject"] == "dest object"
        assert gcs_span.data["gcs"]["sourceBucket"] == "src bucket"
        assert gcs_span.data["gcs"]["sourceObject"] == "src object"

    @patch("requests.Session.request")
    def test_objects_update(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").update()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objects.update"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_default_acls_list(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#objectAccessControls", "items": []},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").default_object_acl.get_entities()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "defaultAcls.list"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"

    @patch("requests.Session.request")
    def test_object_acls_list(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#objectAccessControls", "items": []},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.bucket("test bucket").blob("test object").acl.get_entities()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "objectAcls.list"
        assert gcs_span.data["gcs"]["bucket"] == "test bucket"
        assert gcs_span.data["gcs"]["object"] == "test object"

    @patch("requests.Session.request")
    def test_object_hmac_keys_create(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.create_hmac_key("test@example.com")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "hmacKeys.create"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"

    @patch("requests.Session.request")
    def test_object_hmac_keys_delete(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response()

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            key = storage.hmac_key.HMACKeyMetadata(client, access_id="test key")
            key.state = storage.hmac_key.HMACKeyMetadata.INACTIVE_STATE
            key.delete()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "hmacKeys.delete"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"
        assert gcs_span.data["gcs"]["accessId"] == "test key"

    @patch("requests.Session.request")
    def test_object_hmac_keys_get(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            storage.hmac_key.HMACKeyMetadata(client, access_id="test key").exists()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "hmacKeys.get"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"
        assert gcs_span.data["gcs"]["accessId"] == "test key"

    @patch("requests.Session.request")
    def test_object_hmac_keys_list(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKeysMetadata", "items": []},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            keys = client.list_hmac_keys()

            for _ in keys:
                pass

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "hmacKeys.list"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"

    @patch("requests.Session.request")
    def test_object_hmac_keys_update(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#hmacKey", "metadata": {}, "secret": ""},
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            storage.hmac_key.HMACKeyMetadata(client, access_id="test key").update()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "hmacKeys.update"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"
        assert gcs_span.data["gcs"]["accessId"] == "test key"

    @patch("requests.Session.request")
    def test_object_get_service_account_email(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={
                "email_address": "test@example.com",
                "kind": "storage#serviceAccount",
            },
            status_code=http_client.OK,
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"):
            client.get_service_account_email()

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

        current_span = get_current_span()
        assert not current_span.is_recording()

        gcs_span = spans[0]
        test_span = spans[1]

        self.assertTraceContextPropagated(test_span, gcs_span)

        assert gcs_span.n == "gcs"
        assert gcs_span.k is SpanKind.CLIENT
        assert not gcs_span.ec

        assert gcs_span.data["gcs"]["op"] == "serviceAccount.get"
        assert gcs_span.data["gcs"]["projectId"] == "test-project"

    @patch("requests.Session.request")
    def test_batch_operation(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            _TWO_PART_BATCH_RESPONSE,
            status_code=http_client.OK,
            headers={"content-type": 'multipart/mixed; boundary="DEADBEEF="'},
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        bucket = client.bucket("test-bucket")

        with tracer.start_as_current_span("test"):
            with client.batch():
                for obj in ["obj1", "obj2"]:
                    bucket.delete_blob(obj)

        spans = self.recorder.queued_spans()

        assert len(spans) == 2

    @patch("requests.Session.request")
    def test_execute_with_instana_without_tags(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#buckets", "items": []},
            status_code=http_client.OK,
        )
        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.google.cloud.storage._collect_attributes",
            return_value=None,
        ):
            buckets = client.list_buckets()
            for b in buckets:
                pass
            assert isinstance(buckets, page_iterator.HTTPIterator)

    def test_execute_with_instana_tracing_is_off(self) -> None:
        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.google.cloud.storage.tracing_is_off",
            return_value=True,
        ):
            response = client.list_buckets()
            assert isinstance(response.client, storage.Client)

    @pytest.mark.skipif(
        sys.version_info >= (3, 14),
        reason='Avoiding "Fatal Python error: Segmentation fault"',
    )
    @patch("requests.Session.request")
    def test_download_with_instana_tracing_is_off(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            content=b"CONTENT", status_code=http_client.OK
        )
        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.google.cloud.storage.tracing_is_off",
            return_value=True,
        ):
            response = (
                client.bucket("test bucket")
                .blob("test object")
                .download_to_file(
                    io.BytesIO(),
                    raw_download=True,
                )
            )
            assert not response

    @patch("requests.Session.request")
    def test_upload_with_instana_tracing_is_off(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            json_content={"kind": "storage#object"}, status_code=http_client.OK
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )

        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.google.cloud.storage.tracing_is_off",
            return_value=True,
        ):
            response = (
                client.bucket("test bucket")
                .blob("test object")
                .upload_from_string("CONTENT")
            )
            assert not response

    @patch("requests.Session.request")
    def test_finish_batch_operation_tracing_is_off(self, mock_requests: Mock) -> None:
        mock_requests.return_value = self._mock_response(
            _TWO_PART_BATCH_RESPONSE,
            status_code=http_client.OK,
            headers={"content-type": 'multipart/mixed; boundary="DEADBEEF="'},
        )

        client = self._client(
            credentials=AnonymousCredentials(), project="test-project"
        )
        bucket = client.bucket("test-bucket")

        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.google.cloud.storage.tracing_is_off",
            return_value=True,
        ):
            with client.batch() as batch_response:
                for obj in ["obj1", "obj2"]:
                    bucket.delete_blob(obj)
                assert batch_response

    def _client(self, *args, **kwargs) -> storage.Client:
        # override the HTTP client to bypass the authorization
        kwargs["_http"] = kwargs.get("_http", requests.Session())
        kwargs["_http"].is_mtls = False

        return storage.Client(*args, **kwargs)

    def _mock_response(
        self,
        content=b"",
        status_code=http_client.NO_CONTENT,
        json_content=None,
        headers={},
    ) -> Mock:
        resp = Mock()
        resp.status_code = status_code
        resp.headers = headers
        resp.content = content
        resp.__enter__ = Mock(return_value=resp)
        resp.__exit__ = Mock()

        if json_content is not None:
            if resp.content == b"":
                resp.content = json.dumps(json_content)

            resp.json = Mock(return_value=json_content)

        return resp


_TWO_PART_BATCH_RESPONSE = b"""\
--DEADBEEF=
Content-Type: application/json
Content-ID: <response-1+1>

HTTP/1.1 204 No Content

Content-Type: application/json; charset=UTF-8
Content-Length: 0

--DEADBEEF=
Content-Type: application/json
Content-ID: <response-1+2>

HTTP/1.1 204 No Content

Content-Type: application/json; charset=UTF-8
Content-Length: 0

--DEADBEEF=--
"""



# (c) Copyright IBM Corp. 2024

import os
from tests.apps.bottle_app.app import bottle_server as server
from tests.apps.utils import launch_background_thread

app_thread = None

if not os.environ.get('CASSANDRA_TEST') and app_thread is None:
    app_thread = launch_background_thread(server.serve_forever, "Bottle")


# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from typing import Generator

import pytest
from sqlalchemy import Column, Integer, String, create_engine, text
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import declarative_base, sessionmaker

from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from tests.helpers import testenv

engine = create_engine(
    f"postgresql://{testenv['postgresql_user']}:{testenv['postgresql_pw']}@{testenv['postgresql_host']}:{testenv['postgresql_port']}/{testenv['postgresql_db']}"
)

Session = sessionmaker(bind=engine)
Base = declarative_base()


class StanUser(Base):
    __tablename__ = "churchofstan"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    fullname = Column(String)
    password = Column(String)

    def __repr__(self) -> None:
        return "<User(name='%s', fullname='%s', password='%s')>" % (
            self.name,
            self.fullname,
            self.password,
        )


@pytest.fixture(scope="class")
def db_setup() -> None:
    with tracer.start_as_current_span("metadata") as span:
        Base.metadata.create_all(engine)
        span.end()


stan_user = StanUser(
    name="IAmStan",
    fullname="Stan Robot",
    password="3X}vP66ADoCFT2g?HPvoem2eJh,zWXgd36Rb/{aRq/>7EYy6@EEH4BP(oeXac@mR",
)
stan_user2 = StanUser(
    name="IAmStanToo",
    fullname="Stan Robot 2",
    password="3X}vP66ADoCFT2g?HPvoem2eJh,zWXgd36Rb/{aRq/>7EYy6@EEH4BP(oeXac@mR",
)

sqlalchemy_url = f"postgresql://{testenv['postgresql_host']}:{testenv['postgresql_port']}/{testenv['postgresql_db']}"


@pytest.mark.usefixtures("db_setup")
class TestSQLAlchemy:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.session = Session()
        yield
        """Ensure that allow_exit_as_root has the default value"""
        self.session.close()
        agent.options.allow_exit_as_root = False

    def test_session_add(self) -> None:
        with tracer.start_as_current_span("test"):
            self.session.add(stan_user)
            self.session.commit()

        spans = self.recorder.queued_spans()

        sql_span = spans[0]
        test_span = spans[1]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert sql_span.t == test_span.t

        # Parent relationships
        assert sql_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not sql_span.ec

        # SQLAlchemy span
        assert sql_span.n == "sqlalchemy"
        assert "custom" not in sql_span.data
        assert "sqlalchemy" in sql_span.data

        assert sql_span.data["sqlalchemy"]["eng"] == "postgresql"
        assert sqlalchemy_url == sql_span.data["sqlalchemy"]["url"]
        assert (
            "INSERT INTO churchofstan (name, fullname, password) VALUES (%(name)s, %(fullname)s, %(password)s) RETURNING churchofstan.id"
            == sql_span.data["sqlalchemy"]["sql"]
        )
        assert not sql_span.data["sqlalchemy"]["err"]

        assert sql_span.stack
        assert isinstance(sql_span.stack, list)
        assert len(sql_span.stack) > 0

    def test_session_add_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.session.add(stan_user2)
        self.session.commit()

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        sql_span = spans[0]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Parent relationships
        assert not sql_span.p

        # Error logging
        assert not sql_span.ec

        # SQLAlchemy span
        assert sql_span.n == "sqlalchemy"
        assert "custom" not in sql_span.data
        assert "sqlalchemy" in sql_span.data

        assert sql_span.data["sqlalchemy"]["eng"] == "postgresql"
        assert sqlalchemy_url == sql_span.data["sqlalchemy"]["url"]
        assert (
            "INSERT INTO churchofstan (name, fullname, password) VALUES (%(name)s, %(fullname)s, %(password)s) RETURNING churchofstan.id"
            == sql_span.data["sqlalchemy"]["sql"]
        )
        assert not sql_span.data["sqlalchemy"]["err"]

        assert sql_span.stack
        assert isinstance(sql_span.stack, list)
        assert len(sql_span.stack) > 0

    def test_transaction(self) -> None:
        with tracer.start_as_current_span("test"):
            with engine.begin() as connection:
                connection.execute(text("select 1"))
                connection.execute(
                    text(
                        "select (name, fullname, password) from churchofstan where name='doesntexist'"
                    )
                )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        sql_span0 = spans[0]
        sql_span1 = spans[1]
        test_span = spans[2]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert sql_span0.t == test_span.t
        assert sql_span1.t == test_span.t

        # Parent relationships
        assert sql_span0.p == test_span.s
        assert sql_span1.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not sql_span0.ec
        assert not sql_span1.ec

        # SQLAlchemy span0
        assert sql_span0.n == "sqlalchemy"
        assert "custom" not in sql_span0.data
        assert "sqlalchemy" in sql_span0.data

        assert sql_span0.data["sqlalchemy"]["eng"] == "postgresql"
        assert sqlalchemy_url == sql_span0.data["sqlalchemy"]["url"]
        assert sql_span0.data["sqlalchemy"]["sql"] == "select 1"
        assert not sql_span0.data["sqlalchemy"]["err"]

        assert sql_span0.stack
        assert isinstance(sql_span0.stack, list)
        assert len(sql_span0.stack) > 0

        # SQLAlchemy span1
        assert sql_span1.n == "sqlalchemy"
        assert "custom" not in sql_span1.data
        assert "sqlalchemy" in sql_span1.data

        assert sql_span1.data["sqlalchemy"]["eng"] == "postgresql"
        assert sqlalchemy_url == sql_span1.data["sqlalchemy"]["url"]
        assert (
            "select (name, fullname, password) from churchofstan where name='doesntexist'"
            == sql_span1.data["sqlalchemy"]["sql"]
        )
        assert not sql_span1.data["sqlalchemy"]["err"]

        assert sql_span1.stack
        assert isinstance(sql_span1.stack, list)
        assert len(sql_span1.stack) > 0

    def test_error_logging(self) -> None:
        with tracer.start_as_current_span("test"):
            try:
                self.session.execute(text("htVwGrCwVThisIsInvalidSQLaw4ijXd88"))
                # self.session.commit()
            except Exception:
                pass

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        sql_span = spans[0]
        test_span = spans[1]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert sql_span.t == test_span.t

        # Parent relationships
        assert sql_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert sql_span.ec == 1

        # SQLAlchemy span
        assert sql_span.n == "sqlalchemy"

        assert "custom" not in sql_span.data
        assert "sqlalchemy" in sql_span.data

        assert sql_span.data["sqlalchemy"]["eng"] == "postgresql"
        assert sqlalchemy_url == sql_span.data["sqlalchemy"]["url"]
        assert (
            "htVwGrCwVThisIsInvalidSQLaw4ijXd88" == sql_span.data["sqlalchemy"]["sql"]
        )
        assert (
            'syntax error at or near "htVwGrCwVThisIsInvalidSQLaw4ijXd88'
            in sql_span.data["sqlalchemy"]["err"]
        )
        assert sql_span.stack
        assert isinstance(sql_span.stack, list)
        assert len(sql_span.stack) > 0

    def test_error_before_tracing(self) -> None:
        """Test the scenario, in which instana is loaded,
        but connection fails before tracing begins.
        This is typical in test container scenario,
        where it is "normal" to just start hammering a database container
        which is still starting and not ready to handle requests yet.
        In this scenario it is important that we get
        an sqlalachemy exception, and not something else
        like an AttributeError. Because testcontainer has a logic
        to retry in case of certain sqlalchemy exceptions but it
        can't handle an AttributeError."""
        # https://github.com/instana/python-sensor/issues/362

        current_span = get_current_span()
        assert not current_span.is_recording()

        invalid_connection_url = "postgresql://user1:pwd1@localhost:9999/mydb1"
        with pytest.raises(
            OperationalError,
            match=r"\(psycopg2.OperationalError\) connection .* failed.*",
        ) as context_manager:
            engine = create_engine(invalid_connection_url)
            with engine.connect() as connection:
                (version,) = connection.execute(text("select version()")).fetchone()

        the_exception = context_manager.value
        assert not the_exception.connection_invalidated

    def test_if_not_tracing(self) -> None:
        with engine.begin() as connection:
            connection.execute(text("select 1"))
            connection.execute(
                text(
                    "select (name, fullname, password) from churchofstan where name='doesntexist'"
                )
            )

        current_span = get_current_span()
        assert not current_span.is_recording()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import pytest
import json
from typing import Generator
import boto3
from moto import mock_aws

from instana.singletons import tracer, agent
from tests.helpers import get_first_span_by_filter


class TestLambda:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.mock = mock_aws(config={"lambda": {"use_docker": False}})
        self.mock.start()
        self.lambda_region = "us-east-1"
        self.aws_lambda = boto3.client("lambda", region_name=self.lambda_region)
        self.function_name = "myfunc"
        yield
        # Stop Moto after each test
        self.mock.stop()
        agent.options.allow_exit_as_root = False

    def test_lambda_invoke(self) -> None:
        with tracer.start_as_current_span("test"):
            result = self.aws_lambda.invoke(
                FunctionName=self.function_name,
                Payload=json.dumps({"message": "success"}),
            )

        assert result["StatusCode"] == 200
        result_payload = json.loads(result["Payload"].read().decode("utf-8"))
        assert "message" in result_payload
        assert result_payload["message"] == "success"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "Invoke"
        endpoint = f"https://lambda.{self.lambda_region}.amazonaws.com"
        assert boto_span.data["boto3"]["ep"] == endpoint
        assert boto_span.data["boto3"]["reg"] == self.lambda_region
        assert "FunctionName" in boto_span.data["boto3"]["payload"]
        assert boto_span.data["boto3"]["payload"]["FunctionName"] == self.function_name
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert boto_span.data["http"]["url"] == f"{endpoint}:443/Invoke"

    def test_lambda_invoke_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        result = self.aws_lambda.invoke(
            FunctionName=self.function_name, Payload=json.dumps({"message": "success"})
        )

        assert result["StatusCode"] == 200
        result_payload = json.loads(result["Payload"].read().decode("utf-8"))
        assert "message" in result_payload
        assert result_payload["message"] == "success"

        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        boto_span = spans[0]
        assert boto_span
        assert boto_span.n == "boto3"
        assert not boto_span.p
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "Invoke"
        endpoint = f"https://lambda.{self.lambda_region}.amazonaws.com"
        assert boto_span.data["boto3"]["ep"] == endpoint
        assert boto_span.data["boto3"]["reg"] == self.lambda_region
        assert "FunctionName" in boto_span.data["boto3"]["payload"]
        assert boto_span.data["boto3"]["payload"]["FunctionName"] == self.function_name
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert boto_span.data["http"]["url"] == f"{endpoint}:443/Invoke"

    def test_request_header_capture_before_call(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        # Access the event system on the S3 client
        event_system = self.aws_lambda.meta.events

        request_headers = {"X-Capture-This": "this", "X-Capture-That": "that"}

        # Create a function that adds custom headers
        def add_custom_header_before_call(params, **kwargs):
            params["headers"].update(request_headers)

        # Register the function to before-call event.
        event_system.register(
            "before-call.lambda.Invoke", add_custom_header_before_call
        )

        with tracer.start_as_current_span("test"):
            result = self.aws_lambda.invoke(
                FunctionName=self.function_name,
                Payload=json.dumps({"message": "success"}),
            )

        assert result["StatusCode"] == 200
        result_payload = json.loads(result["Payload"].read().decode("utf-8"))
        assert "message" in result_payload
        assert result_payload["message"] == "success"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "Invoke"
        endpoint = f"https://lambda.{self.lambda_region}.amazonaws.com"
        assert boto_span.data["boto3"]["ep"] == endpoint
        assert boto_span.data["boto3"]["reg"] == self.lambda_region
        assert "FunctionName" in boto_span.data["boto3"]["payload"]
        assert boto_span.data["boto3"]["payload"]["FunctionName"] == self.function_name
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert boto_span.data["http"]["url"] == f"{endpoint}:443/Invoke"

        assert "X-Capture-This" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture_before_sign(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Custom-1", "X-Custom-2"]

        # Access the event system on the S3 client
        event_system = self.aws_lambda.meta.events

        request_headers = {"X-Custom-1": "Value1", "X-Custom-2": "Value2"}

        # Create a function that adds custom headers
        def add_custom_header_before_sign(request, **kwargs):
            for name, value in request_headers.items():
                request.headers.add_header(name, value)

        # Register the function to before-sign event.
        event_system.register_first(
            "before-sign.lambda.Invoke", add_custom_header_before_sign
        )

        with tracer.start_as_current_span("test"):
            result = self.aws_lambda.invoke(
                FunctionName=self.function_name,
                Payload=json.dumps({"message": "success"}),
            )

        assert result["StatusCode"] == 200
        result_payload = json.loads(result["Payload"].read().decode("utf-8"))
        assert "message" in result_payload
        assert result_payload["message"] == "success"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "Invoke"
        endpoint = f"https://lambda.{self.lambda_region}.amazonaws.com"
        assert boto_span.data["boto3"]["ep"] == endpoint
        assert boto_span.data["boto3"]["reg"] == self.lambda_region
        assert "FunctionName" in boto_span.data["boto3"]["payload"]
        assert boto_span.data["boto3"]["payload"]["FunctionName"] == self.function_name
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert boto_span.data["http"]["url"] == f"{endpoint}:443/Invoke"

        assert "X-Custom-1" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-1"] == "Value1"
        assert "X-Custom-2" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Custom-2"] == "Value2"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_response_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        # Access the event system on the S3 client
        event_system = self.aws_lambda.meta.events

        response_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        # Create a function that sets the custom headers in the after-call event.
        def modify_after_call_args(parsed, **kwargs):
            parsed["ResponseMetadata"]["HTTPHeaders"].update(response_headers)

        # Register the function to an event
        event_system.register("after-call.lambda.Invoke", modify_after_call_args)

        with tracer.start_as_current_span("test"):
            result = self.aws_lambda.invoke(
                FunctionName=self.function_name,
                Payload=json.dumps({"message": "success"}),
            )

        assert result["StatusCode"] == 200
        result_payload = json.loads(result["Payload"].read().decode("utf-8"))
        assert "message" in result_payload
        assert result_payload["message"] == "success"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "boto3"
        boto_span = get_first_span_by_filter(spans, filter)
        assert boto_span

        assert boto_span.t == test_span.t
        assert boto_span.p == test_span.s

        assert not test_span.ec
        assert not boto_span.ec

        assert boto_span.data["boto3"]["op"] == "Invoke"
        endpoint = f"https://lambda.{self.lambda_region}.amazonaws.com"
        assert boto_span.data["boto3"]["ep"] == endpoint
        assert boto_span.data["boto3"]["reg"] == self.lambda_region
        assert "FunctionName" in boto_span.data["boto3"]["payload"]
        assert boto_span.data["boto3"]["payload"]["FunctionName"] == self.function_name
        assert boto_span.data["http"]["status"] == 200
        assert boto_span.data["http"]["method"] == "POST"
        assert boto_span.data["http"]["url"] == f"{endpoint}:443/Invoke"

        assert "X-Capture-This-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in boto_span.data["http"]["header"]
        assert boto_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


from instana.log import logger
from instana.propagators.base_propagator import BasePropagator
from instana.util.ids import define_server_timing, hex_id_limited

from opentelemetry.trace.span import format_span_id

class HTTPPropagator(BasePropagator):
    """
    Instana Propagator for Format.HTTP_HEADERS.

    The HTTP_HEADERS format deals with key-values with string to string mapping.
    The character set should be restricted to HTTP compatible.
    """

    def __init__(self):
        super(HTTPPropagator, self).__init__()

    def inject(self, span_context, carrier, disable_w3c_trace_context=False):
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        dictionary_carrier = self.extract_headers_dict(carrier)
        if dictionary_carrier:
            # Suppression `level` made in the child context or in the parent context
            # has priority over any non-suppressed `level` setting
            child_level = int(self.extract_instana_headers(dictionary_carrier)[2] or "1")
            span_context.level = min(child_level, span_context.level)

        serializable_level = str(span_context.level)

        if disable_w3c_trace_context:
            traceparent, tracestate = [None] * 2
        else:
            traceparent, tracestate = self._get_participating_trace_context(span_context)

        def inject_key_value(carrier, key, value):
            if isinstance(carrier, list):
                carrier.append((key, value))
            elif isinstance(carrier, dict) or '__setitem__' in dir(carrier):
                carrier[key] = value
            else:
                raise Exception("Unsupported carrier type", type(carrier))

        try:
            inject_key_value(carrier, self.HEADER_KEY_L, serializable_level)

            if traceparent:
                inject_key_value(carrier, self.HEADER_KEY_TRACEPARENT, traceparent)
            if tracestate:
                inject_key_value(carrier, self.HEADER_KEY_TRACESTATE, tracestate)

            if span_context.suppression:
                return

            inject_key_value(carrier, self.HEADER_KEY_T, hex_id_limited(trace_id))
            inject_key_value(carrier, self.HEADER_KEY_S, format_span_id(span_id))
            inject_key_value(
                carrier, self.HEADER_KEY_SERVER_TIMING, define_server_timing(trace_id)
            )

        except Exception:
            logger.debug("inject error:", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import random
import time
from typing import TYPE_CHECKING, Union

from instana.autoprofile.schedule import delay, schedule
from instana.log import logger

if TYPE_CHECKING:
    from instana.autoprofile.profiler import Profiler
    from instana.autoprofile.samplers.allocation_sampler import AllocationSampler
    from instana.autoprofile.samplers.block_sampler import BlockSampler
    from instana.autoprofile.samplers.cpu_sampler import CPUSampler


class SamplerConfig(object):
    def __init__(self) -> None:
        self.log_prefix = None
        self.max_profile_duration = None
        self.max_span_duration = None
        self.span_interval = None
        self.report_interval = None


class SamplerScheduler:
    def __init__(
        self,
        profiler: "Profiler",
        sampler: Union["AllocationSampler", "BlockSampler", "CPUSampler"],
        config: SamplerConfig,
    ) -> None:
        self.profiler = profiler
        self.sampler = sampler
        self.config = config
        self.started = False
        self.span_timer = None
        self.span_timeout = None
        self.random_timer = None
        self.report_timer = None
        self.profile_start_ts = None
        self.profile_duration = None
        self.span_active = False
        self.span_start_ts = None
        self.span_count = 0

    def setup(self) -> None:
        self.sampler.setup()

    def start(self) -> None:
        if not self.sampler.ready:
            return

        if self.started:
            return

        self.started = True
        self.reset()

        def random_delay() -> None:
            timeout = random.randint(
                0, round(self.config.span_interval - self.config.max_span_duration)
            )
            self.random_timer = delay(timeout, self.start_profiling)

        if not self.profiler.get_option("disable_timers"):
            self.span_timer = schedule(0, self.config.span_interval, random_delay)
            self.report_timer = schedule(
                self.config.report_interval, self.config.report_interval, self.report
            )

    def stop(self) -> None:
        if not self.started:
            return

        self.started = False

        if self.span_timer:
            self.span_timer.cancel()
            self.span_timer = None

        if self.random_timer:
            self.random_timer.cancel()
            self.random_timer = None

        if self.report_timer:
            self.report_timer.cancel()
            self.report_timer = None

        self.stop_profiling()

    def destroy(self) -> None:
        self.sampler.destroy()

    def reset(self) -> None:
        self.sampler.reset()
        self.profile_start_ts = time.time()
        self.profile_duration = 0
        self.span_count = 0

    def start_profiling(self) -> bool:
        if not self.started:
            return False

        if self.profile_duration > self.config.max_profile_duration:
            logger.debug(f"{self.config.log_prefix}: max profiling duration reached.")
            return False

        if self.span_count > self.config.max_span_count:
            logger.debug(f"{self.config.log_prefix}: max recording count reached.")
            return False

        if self.profiler.sampler_active:
            logger.debug(f"{self.config.log_prefix}: sampler lock exists.")
            return False

        self.profiler.sampler_active = True
        logger.debug(f"{self.config.log_prefix}: started.")

        try:
            self.sampler.start_sampler()
        except Exception:
            self.profiler.sampler_active = False
            logger.error("Error starting profiling", exc_info=True)
            return False

        self.span_timeout = delay(self.config.max_span_duration, self.stop_profiling)

        self.span_active = True
        self.span_start_ts = time.time()
        self.span_count += 1

        return True

    def stop_profiling(self) -> None:
        if not self.span_active:
            return

        self.span_active = False

        try:
            self.profile_duration = (
                self.profile_duration + time.time() - self.span_start_ts
            )
            self.sampler.stop_sampler()
        except Exception:
            logger.error("Error stopping profiling", exc_info=True)

        self.profiler.sampler_active = False

        if self.span_timeout:
            self.span_timeout.cancel()

        logger.debug(f"{self.config.log_prefix}: stopped.")

    def report(self) -> None:
        if not self.started:
            return

        if self.profile_duration == 0:
            return

        if self.profile_start_ts > time.time() - self.config.report_interval:
            return
        elif self.profile_start_ts < time.time() - 2 * self.config.report_interval:
            self.reset()
            return

        profile = self.sampler.build_profile(
            to_millis(self.profile_duration),
            to_millis(time.time() - self.profile_start_ts),
        )

        if self.profiler.agent.can_send():
            if self.profiler.agent.announce_data.pid:
                profile.process_id = str(self.profiler.agent.announce_data.pid)

            self.profiler.agent.collector.profile_queue.put(profile.to_dict())

            logger.debug(f"{self.config.log_prefix}: reporting profile:")
        else:
            logger.debug(
                f"{self.config.log_prefix}: not reporting profile, agent not ready"
            )

        self.reset()


def to_millis(t: int) -> int:
    return int(round(t * 1000))



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


try:
    import flask

    # `signals_available` indicates whether the Flask process is running with or without blinker support:
    # https://pypi.org/project/blinker/
    #
    # Blinker support is preferred but we do the best we can when it's not available.
    #
    if hasattr(flask.signals, 'signals_available'):
        from flask.signals import signals_available
    else:
        # Beginning from 2.3.0 as stated in the notes
        # https://flask.palletsprojects.com/en/2.3.x/changes/#version-2-3-0
        # "Signals are always available. blinker>=1.6.2 is a required dependency.
        # The signals_available attribute is deprecated. #5056"
        signals_available = True

    from instana.instrumentation.flask import common

    if signals_available is True:
        import instana.instrumentation.flask.with_blinker
    else:
        import instana.instrumentation.flask.vanilla
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import asyncio
from typing import Any, Dict, Generator, Optional

import aiohttp
import pytest

import tests.apps.flask_app  # noqa: F401
from instana.configurator import config
from instana.singletons import tracer
from tests.helpers import testenv


class TestAsyncio:
    async def fetch(
        self,
        session: aiohttp.ClientSession,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ):
        try:
            async with session.get(url, headers=headers, params=params) as response:
                return response
        except aiohttp.web_exceptions.HTTPException:
            pass

    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # New event loop for every test
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        # Restore default
        config["asyncio_task_context_propagation"]["enabled"] = False
        yield
        # teardown
        # Close the loop if running
        if self.loop.is_running():
            self.loop.close()

    def test_ensure_future_with_context(self) -> None:
        async def run_later(msg="Hello"):
            async with aiohttp.ClientSession() as session:
                return await self.fetch(session, testenv["flask_server"] + "/")

        async def test():
            with tracer.start_as_current_span("test"):
                asyncio.ensure_future(run_later("Hello OTel"))
            await asyncio.sleep(0.5)

        # Override default task context propagation
        config["asyncio_task_context_propagation"]["enabled"] = True

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        test_span = spans[0]
        wsgi_span = spans[1]
        aioclient_span = spans[2]

        assert test_span.t == wsgi_span.t
        assert aioclient_span.t == test_span.t

        assert not test_span.p
        assert wsgi_span.p == aioclient_span.s
        assert aioclient_span.p == test_span.s

    def test_ensure_future_without_context(self) -> None:
        async def run_later(msg="Hello"):
            async with aiohttp.ClientSession() as session:
                return await self.fetch(session, testenv["flask_server"] + "/")

        async def test():
            with tracer.start_as_current_span("test"):
                asyncio.ensure_future(run_later("Hello OTel"))
            await asyncio.sleep(0.5)

        self.loop.run_until_complete(test())

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].n == "sdk"
        assert spans[1].n == "wsgi"

        # Without the context propagated, we should get two separate traces
        assert spans[0].t != spans[1].t

    if hasattr(asyncio, "create_task"):

        def test_create_task_with_context(self) -> None:
            async def run_later(msg="Hello"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/")

            async def test():
                with tracer.start_as_current_span("test"):
                    asyncio.create_task(run_later("Hello OTel"))
                await asyncio.sleep(0.5)

            # Override default task context propagation
            config["asyncio_task_context_propagation"]["enabled"] = True

            self.loop.run_until_complete(test())

            spans = self.recorder.queued_spans()
            assert len(spans) == 3

            test_span = spans[0]
            wsgi_span = spans[1]
            aioclient_span = spans[2]

            assert wsgi_span.t == test_span.t
            assert aioclient_span.t == test_span.t

            assert not test_span.p
            assert wsgi_span.p == aioclient_span.s
            assert aioclient_span.p == test_span.s

        def test_create_task_without_context(self) -> None:
            async def run_later(msg="Hello"):
                async with aiohttp.ClientSession() as session:
                    return await self.fetch(session, testenv["flask_server"] + "/")

            async def test():
                with tracer.start_as_current_span("test"):
                    asyncio.create_task(run_later("Hello OTel"))
                await asyncio.sleep(0.5)

            self.loop.run_until_complete(test())

            spans = self.recorder.queued_spans()

            assert len(spans) == 2
            assert spans[0].n == "sdk"
            assert spans[1].n == "wsgi"

            # Without the context propagated, we should get two separate traces
            assert spans[0].t != spans[1].t



# (c) Copyright IBM Corp. 2024

from typing import Generator

import pytest

from instana.sampling import InstanaSampler, SamplingPolicy


class TestInstanaSampler:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.sampler = InstanaSampler()
        yield
        self.sampler = None

    def test_sampling_policy(self) -> None:
        assert self.sampler._sampled == SamplingPolicy.DROP
        assert self.sampler._sampled.name == "DROP"
        assert self.sampler._sampled.value == 0

    def test_sampler(self) -> None:
        assert not self.sampler.sampled()



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

"""
Instrumentation for Sanic
https://sanicframework.org/en/
"""

try:
    import sanic
    from instana.log import logger

    if not (hasattr(sanic, "__version__") and sanic.__version__ >= "19.9.0"):
        logger.debug(
            "Instana supports Sanic package versions 19.9.0 and newer.  Skipping."
        )
        raise ImportError

    import wrapt
    from typing import Callable, Tuple, Dict, Any
    from sanic.exceptions import SanicException

    from opentelemetry import context, trace
    from opentelemetry.trace import SpanKind
    from opentelemetry.semconv.trace import SpanAttributes

    from instana.singletons import tracer, agent
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import extract_custom_headers
    from instana.propagators.format import Format

    from sanic.request import Request
    from sanic.response import HTTPResponse

    @wrapt.patch_function_wrapper("sanic.app", "Sanic.__init__")
    def init_with_instana(
        wrapped: Callable[..., sanic.app.Sanic.__init__],
        instance: sanic.app.Sanic,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        wrapped(*args, **kwargs)
        app = instance

        @app.middleware("request")
        def request_with_instana(request: Request) -> None:
            try:
                if "http" not in request.scheme:
                    return

                headers = request.headers.copy()
                parent_context = tracer.extract(Format.HTTP_HEADERS, headers)

                span = tracer.start_span("asgi", span_context=parent_context)
                request.ctx.span = span

                ctx = trace.set_span_in_context(span)
                token = context.attach(ctx)
                request.ctx.token = token

                span.set_attribute("span.kind", SpanKind.SERVER)
                span.set_attribute("http.path", request.path)
                span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)
                span.set_attribute(SpanAttributes.HTTP_HOST, request.host)
                if hasattr(request, "url"):
                    span.set_attribute(SpanAttributes.HTTP_URL, request.url)

                query = request.query_string

                if isinstance(query, (str, bytes)) and len(query):
                    if isinstance(query, bytes):
                        query = query.decode("utf-8")
                    scrubbed_params = strip_secrets_from_query(
                        query, agent.options.secrets_matcher, agent.options.secrets_list
                    )
                    span.set_attribute("http.params", scrubbed_params)

                extract_custom_headers(span, headers)
                if hasattr(request, "uri_template") and request.uri_template:
                    span.set_attribute("http.path_tpl", request.uri_template)
            except Exception:
                logger.debug("request_with_instana: ", exc_info=True)

        @app.exception(Exception)
        def exception_with_instana(request: Request, exception: Exception) -> None:
            try:
                if not hasattr(request.ctx, "span"):  # pragma: no cover
                    return
                span = request.ctx.span

                if isinstance(exception, SanicException):
                    # Handle Sanic-specific exceptions
                    status_code = exception.status_code
                    message = str(exception)

                    if all([span, status_code, message]) and 500 <= status_code:
                        span.set_attribute("http.error", message)
            except Exception:
                logger.debug("exception_with_instana: ", exc_info=True)

        @app.middleware("response")
        def response_with_instana(request: Request, response: HTTPResponse) -> None:
            try:
                if not hasattr(request.ctx, "span"):  # pragma: no cover
                    return
                span = request.ctx.span

                status_code = response.status
                if status_code:
                    if int(status_code) >= 500:
                        span.mark_as_errored()
                    span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)

                if hasattr(response, "headers"):
                    extract_custom_headers(span, response.headers)
                    tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)

                if span.is_recording():
                    span.end()
                request.ctx.span = None

                if request.ctx.token:
                    context.detach(request.ctx.token)
                    request.ctx.token = None
            except Exception:
                logger.debug("response_with_instana: ", exc_info=True)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2024

from typing import TYPE_CHECKING, Type
import six

from instana.log import logger
from instana.util import DictionaryOfStan
from instana.span.kind import ENTRY_SPANS

if TYPE_CHECKING:
    from opentelemetry.trace import Span


class BaseSpan(object):
    sy = None

    def __str__(self) -> str:
        return "BaseSpan(%s)" % self.__dict__.__str__()

    def __repr__(self) -> str:
        return self.__dict__.__str__()

    def __init__(self, span: Type["Span"], source, **kwargs) -> None:
        # pylint: disable=invalid-name
        self.t = span.context.trace_id
        self.p = span.parent_id
        self.s = span.context.span_id
        self.ts = round(span.start_time / 10**6)
        self.d = round(span.duration / 10**6) if span.duration else None
        self.f = source
        self.ec = span.attributes.pop("ec", None)
        self.data = DictionaryOfStan()
        self.stack = span.stack

        if span.synthetic is True and span.name in ENTRY_SPANS:
            self.sy = span.synthetic

        self.__dict__.update(kwargs)

    def _populate_extra_span_attributes(self, span) -> None:
        if span.context.trace_parent:
            self.tp = span.context.trace_parent
        if span.context.instana_ancestor:
            self.ia = span.context.instana_ancestor
        if span.context.long_trace_id:
            self.lt = span.context.long_trace_id
        if span.context.correlation_type:
            self.crtp = span.context.correlation_type
        if span.context.correlation_id:
            self.crid = span.context.correlation_id

    def _validate_attributes(self, attributes):
        """
        This method will loop through a set of attributes to validate each key and value.

        :param attributes: dict of attributes
        :return: dict - a filtered set of attributes
        """
        filtered_attributes = DictionaryOfStan()
        for key in attributes.keys():
            validated_key, validated_value = self._validate_attribute(
                key, attributes[key]
            )
            if validated_key is not None and validated_value is not None:
                filtered_attributes[validated_key] = validated_value
        return filtered_attributes

    def _validate_attribute(self, key, value):
        """
        This method will assure that <key> and <value> are valid to set as a attribute.
        If <value> fails the check, an attempt will be made to convert it into
        something useful.

        On check failure, this method will return None values indicating that the attribute is
        not valid and could not be converted into something useful

        :param key: The attribute key
        :param value: The attribute value
        :return: Tuple (key, value)
        """
        validated_key = None
        validated_value = None

        try:
            # Attribute keys must be some type of text or string type
            if isinstance(key, (six.text_type, six.string_types)):
                validated_key = key[0:1024]  # Max key length of 1024 characters

                if isinstance(
                    value,
                    (bool, float, int, list, dict, six.text_type, six.string_types),
                ):
                    validated_value = value
                else:
                    validated_value = self._convert_attribute_value(value)
            else:
                logger.debug(
                    "(non-fatal) attribute names must be strings. attribute discarded for %s",
                    type(key),
                )
        except Exception:
            logger.debug("instana.span._validate_attribute: ", exc_info=True)

        return (validated_key, validated_value)

    def _convert_attribute_value(self, value):
        final_value = None

        try:
            final_value = repr(value)
        except Exception:
            final_value = (
                "(non-fatal) span.set_attribute: values must be one of these types: bool, float, int, list, "
                "set, str or alternatively support 'repr'. attribute discarded"
            )
            logger.debug(final_value, exc_info=True)
            return None
        return final_value



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016


import os
import re
import time
import traceback
from contextlib import contextmanager
from typing import TYPE_CHECKING, Iterator, Mapping, Optional, Type, Union

from opentelemetry.context.context import Context
from opentelemetry.trace import (
    SpanKind,
    TraceFlags,
    Tracer,
    TracerProvider,
    _Links,
    use_span,
)
from opentelemetry.util import types

from instana.agent.host import HostAgent
from instana.log import logger
from instana.propagators.binary_propagator import BinaryPropagator
from instana.propagators.exceptions import UnsupportedFormatException
from instana.propagators.format import Format
from instana.propagators.http_propagator import HTTPPropagator
from instana.propagators.text_propagator import TextPropagator
from instana.recorder import StanRecorder
from instana.sampling import InstanaSampler, Sampler
from instana.span.kind import EXIT_SPANS
from instana.span.span import InstanaSpan, get_current_span
from instana.span_context import SpanContext
from instana.util.ids import generate_id

if TYPE_CHECKING:
    from instana.agent.base import BaseAgent
    from instana.propagators.base_propagator import BasePropagator, CarrierT


class InstanaTracerProvider(TracerProvider):
    def __init__(
        self,
        sampler: Optional[Sampler] = None,
        span_processor: Optional[StanRecorder] = None,
        exporter: Optional[Type["BaseAgent"]] = None,
    ) -> None:
        self.sampler = sampler or InstanaSampler()
        self._span_processor = span_processor or StanRecorder()
        self._exporter = exporter or HostAgent()
        self._propagators = {}
        self._propagators[Format.HTTP_HEADERS] = HTTPPropagator()
        self._propagators[Format.TEXT_MAP] = TextPropagator()
        self._propagators[Format.BINARY] = BinaryPropagator()

    def get_tracer(
        self,
        instrumenting_module_name: str,
        instrumenting_library_version: Optional[str] = None,
        schema_url: Optional[str] = None,
        attributes: Optional[types.Attributes] = None,
    ) -> Tracer:
        if not instrumenting_module_name:  # Reject empty strings too.
            instrumenting_module_name = ""
            logger.error("get_tracer called with missing module name.")

        return InstanaTracer(
            self.sampler,
            self._span_processor,
            self._exporter,
            self._propagators,
        )

    def add_span_processor(
        self,
        span_processor: StanRecorder,
    ) -> None:
        """Registers a new SpanProcessor for the TracerProvider."""
        self._span_processor = span_processor


class InstanaTracer(Tracer):
    """Handles :class:`InstanaSpan` creation and in-process context propagation.

    This class provides methods for manipulating the context, creating spans,
    and controlling spans' lifecycles.
    """

    def __init__(
        self,
        sampler: Sampler,
        span_processor: StanRecorder,
        exporter: Type["BaseAgent"],
        propagators: Mapping[str, Type["BasePropagator"]],
    ) -> None:
        self._sampler = sampler
        self._span_processor = span_processor
        self._exporter = exporter
        self._propagators = propagators

    @property
    def span_processor(self) -> Optional[StanRecorder]:
        return self._span_processor

    @property
    def exporter(self) -> Optional[Type["BaseAgent"]]:
        return self._exporter

    def start_span(
        self,
        name: str,
        span_context: Optional[SpanContext] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: _Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
    ) -> InstanaSpan:
        parent_context = span_context if span_context else get_current_span().get_span_context()

        if parent_context and not isinstance(parent_context, SpanContext):
            raise TypeError("parent_context must be an Instana SpanContext or None.")

        span_context = self._create_span_context(parent_context)
        span = InstanaSpan(
            name,
            span_context,
            self._span_processor,
            parent_id=(None if parent_context is None else parent_context.span_id),
            start_time=(time.time_ns() if start_time is None else start_time),
            attributes=attributes,
            # events: Sequence[Event] = None,
        )

        if name in EXIT_SPANS:
            self._add_stack(span)

        return span

    @contextmanager
    def start_as_current_span(
        self,
        name: str,
        span_context: Optional[SpanContext] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: _Links = None,
        start_time: Optional[int] = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ) -> Iterator[InstanaSpan]:
        span = self.start_span(
            name=name,
            span_context=span_context,
            kind=kind,
            attributes=attributes,
            links=links,
            start_time=start_time,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        )
        with use_span(
            span,
            end_on_exit=end_on_exit,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        ) as span:
            yield span

    def _add_stack(self, span: InstanaSpan, limit: Optional[int] = 30) -> None:
        """
        Adds a backtrace to <span>.  The default length limit for
        stack traces is 30 frames.  A hard limit of 40 frames is enforced.
        """
        try:
            sanitized_stack = []
            if limit > 40:
                limit = 40

            trace_back = traceback.extract_stack()
            trace_back.reverse()
            for frame in trace_back:
                # Exclude Instana frames unless we're in dev mode
                if "INSTANA_DEBUG" not in os.environ:
                    if re_tracer_frame.search(frame[0]) is not None:
                        continue

                    if re_with_stan_frame.search(frame[2]) is not None:
                        continue

                sanitized_stack.append({"c": frame[0], "n": frame[1], "m": frame[2]})

            if len(sanitized_stack) > limit:
                # (limit * -1) gives us negative form of <limit> used for
                # slicing from the end of the list. e.g. stack[-30:]
                span.stack = sanitized_stack[(limit * -1) :]
            else:
                span.stack = sanitized_stack
        except Exception:
            # No fail
            pass

    def _create_span_context(self, parent_context: SpanContext) -> SpanContext:
        """Creates a new SpanContext based on the given parent context."""

        if parent_context and parent_context.is_valid:
            trace_id = parent_context.trace_id
            span_id = generate_id()
            trace_flags = parent_context.trace_flags
            is_remote = parent_context.is_remote
        else:
            trace_id = span_id = generate_id()
            trace_flags = TraceFlags(self._sampler.sampled())
            is_remote = False

        span_context = SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            trace_flags=trace_flags,
            is_remote=is_remote,
            level=(parent_context.level if parent_context else 1),
            synthetic=(parent_context.synthetic if parent_context else False),
            trace_parent=(parent_context.trace_parent if parent_context else None),
            instana_ancestor=(parent_context.instana_ancestor if parent_context else None),
            long_trace_id=(parent_context.long_trace_id if parent_context else None),
            correlation_type=(parent_context.correlation_type if parent_context else None),
            correlation_id=(parent_context.correlation_id if parent_context else None),
            traceparent=(parent_context.traceparent if parent_context else None),
            tracestate=(parent_context.tracestate if parent_context else None),
        )

        return span_context

    def inject(
        self,
        span_context: SpanContext,
        format: Union[Format.BINARY, Format.HTTP_HEADERS, Format.TEXT_MAP],
        carrier: "CarrierT",
        disable_w3c_trace_context: bool = False,
    ) -> Optional["CarrierT"]:
        if format in self._propagators:
            return self._propagators[format].inject(
                span_context, carrier, disable_w3c_trace_context
            )

        raise UnsupportedFormatException()

    def extract(
        self,
        format: Union[Format.BINARY, Format.HTTP_HEADERS, Format.TEXT_MAP],
        carrier: "CarrierT",
        disable_w3c_trace_context: bool = False,
    ) -> Optional[Context]:
        if format in self._propagators:
            return self._propagators[format].extract(carrier, disable_w3c_trace_context)

        raise UnsupportedFormatException()


# Used by __add_stack
re_tracer_frame = re.compile(r"/instana/.*\.py$")
re_with_stan_frame = re.compile("with_instana")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

import tests.apps.grpc_server.stan_pb2 as stan__pb2

GRPC_GENERATED_VERSION = "1.67.1"
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + " but the generated code in stan_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
    )


class StanStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.OneQuestionOneResponse = channel.unary_unary(
            "/stan.Stan/OneQuestionOneResponse",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )
        self.ManyQuestionsOneResponse = channel.stream_unary(
            "/stan.Stan/ManyQuestionsOneResponse",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )
        self.OneQuestionManyResponses = channel.unary_stream(
            "/stan.Stan/OneQuestionManyResponses",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )
        self.ManyQuestionsManyReponses = channel.stream_stream(
            "/stan.Stan/ManyQuestionsManyReponses",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )
        self.OneQuestionOneErrorResponse = channel.unary_unary(
            "/stan.Stan/OneQuestionOneErrorResponse",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )
        self.OneErroredQuestionOneResponse = channel.unary_unary(
            "/stan.Stan/OneErroredQuestionOneResponse",
            request_serializer=stan__pb2.QuestionRequest.SerializeToString,
            response_deserializer=stan__pb2.QuestionResponse.FromString,
            _registered_method=True,
        )


class StanServicer(object):
    """Missing associated documentation comment in .proto file."""

    def OneQuestionOneResponse(self, request, context):
        """Unary"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ManyQuestionsOneResponse(self, request_iterator, context):
        """Streaming"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def OneQuestionManyResponses(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ManyQuestionsManyReponses(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def OneQuestionOneErrorResponse(self, request, context):
        """Error Testing"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def OneErroredQuestionOneResponse(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_StanServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "OneQuestionOneResponse": grpc.unary_unary_rpc_method_handler(
            servicer.OneQuestionOneResponse,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
        "ManyQuestionsOneResponse": grpc.stream_unary_rpc_method_handler(
            servicer.ManyQuestionsOneResponse,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
        "OneQuestionManyResponses": grpc.unary_stream_rpc_method_handler(
            servicer.OneQuestionManyResponses,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
        "ManyQuestionsManyReponses": grpc.stream_stream_rpc_method_handler(
            servicer.ManyQuestionsManyReponses,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
        "OneQuestionOneErrorResponse": grpc.unary_unary_rpc_method_handler(
            servicer.OneQuestionOneErrorResponse,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
        "OneErroredQuestionOneResponse": grpc.unary_unary_rpc_method_handler(
            servicer.OneErroredQuestionOneResponse,
            request_deserializer=stan__pb2.QuestionRequest.FromString,
            response_serializer=stan__pb2.QuestionResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "stan.Stan", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))



# (c) Copyright IBM Corp. 2024

import time
from typing import Generator

import pytest
from opentelemetry.trace.status import Status, StatusCode

from instana.span.readable_span import Event, ReadableSpan
from instana.span_context import SpanContext


class TestReadableSpan:
    @pytest.fixture(autouse=True)
    def _resource(
        self,
    ) -> Generator[None, None, None]:
        self.span = None
        yield

    def test_readablespan(
        self,
        span_context: SpanContext,
        trace_id: int,
        span_id: int,
    ) -> None:
        span_name = "test-span"
        timestamp = time.time_ns()
        self.span = ReadableSpan(span_name, span_context)

        assert self.span is not None
        assert isinstance(self.span, ReadableSpan)
        assert self.span.name == span_name

        span_context = self.span.context
        assert isinstance(span_context, SpanContext)
        assert span_context.trace_id == trace_id
        assert span_context.span_id == span_id

        assert self.span.start_time
        assert isinstance(self.span.start_time, int)
        assert self.span.start_time > timestamp
        assert not self.span.end_time
        assert not self.span.attributes
        assert not self.span.events
        assert not self.span.parent_id
        assert not self.span.duration
        assert self.span.status

        assert not self.span.stack
        assert self.span.synthetic is False

    def test_readablespan_with_params(
        self,
        span_context: SpanContext,
    ) -> None:
        span_name = "test-span"
        parent_id = "123456789"
        start_time = time.time_ns()
        end_time = time.time_ns()
        attributes = {"key": "value"}
        event_name = "event"
        events = [Event(event_name, attributes, start_time)]
        status = Status(StatusCode.OK)
        stack = ["span-1", "span-2"]
        self.span = ReadableSpan(
            span_name,
            span_context,
            parent_id,
            start_time,
            end_time,
            attributes,
            events,
            status,
            stack,
        )

        assert self.span.name == span_name
        assert self.span.parent_id == parent_id
        assert self.span.start_time == start_time
        assert self.span.end_time == end_time
        assert self.span.attributes == attributes
        assert self.span.events == events
        assert self.span.status == status
        assert self.span.duration == end_time - start_time
        assert self.span.stack == stack
